from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, TYPE_CHECKING

from pydantic import BaseModel, ConfigDict, Field


from prodsys.simulation import router, sim, source
from prodsys.models.product_data import ProductData
from prodsys.models.source_data import SourceData


if TYPE_CHECKING:
    from prodsys.factories import (
        resource_factory,
        queue_factory,
        time_model_factory,
        sink_factory,
    )
    from prodsys.adapters import adapter


class SourceFactory(BaseModel):
    """
    Factory class that creates and stores `prodsys.simulation` source objects based on the given source data according to `prodsys.models.SourceData`.

    Args:
        env (sim.Environment): prodsys simulation environment.
        product_factory (product_factory.ProductFactory): Factory that creates product objects.
        time_model_factory (time_model_factory.TimeModelFactory): Factory that creates time model objects.
        queue_factory (queue_factory.QueueFactory): Factory that creates queue objects.
        resource_factory (resource_factory.ResourceFactory): Factory that creates resource objects.
        sink_factory (sink_factory.SinkFactory): Factory that creates sink objects.
    """
    env: sim.Environment
    product_factory: product_factory.ProductFactory
    time_model_factory: time_model_factory.TimeModelFactory
    queue_factory: queue_factory.QueueFactory
    resource_factory: resource_factory.ResourceFactory
    auxiliary_factory: auxiliary_factory.AuxiliaryFactory
    sink_factory: sink_factory.SinkFactory

    product_data: List[ProductData] = Field(
        default_factory=list, init=False
    )
    sources: List[source.Source] = Field(default_factory=list, init=False)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def create_sources(self, adapter: adapter.ProductionSystemAdapter):
        """
        Creates source objects based on the given adapter.

        Args:
            adapter (adapter.ProductionSystemAdapter): Adapter that contains the source data.
        """
        for values in adapter.source_data:
            for product_d in adapter.product_data:
                if product_d.product_type == values.product_type:
                    self.add_source(values, product_d)

    def get_router(self, routing_heuristic: str):
        return router.Router(
            self.resource_factory,
            self.sink_factory,
            self.auxiliary_factory,
            router.ROUTING_HEURISTIC[routing_heuristic],
        )

    def add_source(
        self,
        source_data: SourceData,
        product_data_of_source: ProductData,
    ):
        router = self.get_router(source_data.routing_heuristic)

        time_model = self.time_model_factory.get_time_model(source_data.time_model_id)

        source_object = source.Source(
            env=self.env,
            data=source_data,
            product_data=product_data_of_source,
            product_factory=self.product_factory,
            time_model=time_model,
            router=router,
        )
        self.add_queues_to_source(source_object, source_data.output_queues)
        self.sources.append(source_object)

    def add_queues_to_source(self, source: source.Source, values: List[str]):
        output_queues = self.queue_factory.get_queues(values)
        source.add_output_queues(output_queues)

    def start_sources(self):
        """
        Starts the processes of all source objects, i.e. initializes the simulation.
        """
        for _source in self.sources:
            _source.start_source()

    def get_source(self, ID: str) -> source.Source:
        """
        Returns a source object with the given ID.

        Args:
            ID (str): ID of the source object.

        Returns:
            source.Source: Source object with the given ID.
        """
        return [s for s in self.sources if s.data.ID == ID].pop()

    def get_sources(self, IDs: List[str]) -> List[source.Source]:
        """
        Method returns a list of source objects with the given IDs.

        Args:
            IDs (List[str]): List of IDs that is used to sort the source objects.

        Returns:
            List[source.Source]: List of source objects with the given IDs.
        """
        return [s for s in self.sources if s.data.ID in IDs]

    def get_sources_with_product_type(self, __product_type: str) -> List[source.Source]:
        """
        Method returns a list of source objects with the given product type.

        Args:
            __product_type (str): Product type that is used to sort the source objects. 

        Returns:
            List[source.Source]: List of source objects with the given product type.
        """
        return [s for s in self.sources if __product_type == s.data.product_type]


from prodsys.factories import (
    product_factory,
    resource_factory,
    queue_factory,
    time_model_factory,
    auxiliary_factory,
    sink_factory,
)



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Optional, Union, List, Tuple, Union

if TYPE_CHECKING:
    from prodsys.simulation.product import Product, Locatable
    from prodsys.simulation.process import PROCESS_UNION, TransportProcess, LinkTransportProcess
    from prodsys.simulation.resources import Resource, TransportResource
    from prodsys.simulation.sink import Sink
    from prodsys.simulation.auxiliary import Auxiliary



class Request:
    """
    Class to represents requests of a product for a process to be executed by a resource.

    Args:
        process (process.PROCESS_UNION): The process.
        product (product.Product): The product.
        resource (resources.Resource): The resource.
    """
    def __init__(
        self,
        process: PROCESS_UNION,
        product: Product,
        resource: Resource
    ):
        self.process = process
        self.product = product
        self.resource = resource


    def set_process(self, process: PROCESS_UNION):
        """
        Sets the process of the request.

        Args:
            process (process.PROCESS_UNION): The process.
        """
        self.process = process
        # TODO: maybe do some special handling of compound processes here


    def get_process(self) -> PROCESS_UNION:
        """
        Returns the process or the capability process of the request 

        Returns:
            process.PROCESS_UNION: The process.
        """
        return self.process

    def get_product(self) -> Product:
        """
        Returns the product of the request.

        Returns:
            product.Product: The product.
        """
        return self.product

    def get_resource(self) -> Resource:
        """
        Returns the resource of the request.

        Returns:
            resources.Resource: The resource.
        """
        return self.resource


class SinkRequest(Request):
    """
    Class to represents requests of a product for a storage in a sink to be executed by a resource.

    Args:
        Request (_type_): _description_
    """
    def __init__(
        self,
        product: Product,
        sink: Sink
    ):
        self.resource = sink
        self.product = product
        self.process = None
    
class AuxiliaryRequest(Request):
    """
    Represents an auxiliary request in the simulation. The request is associated with an auxiliary which needs to be transported

    Attributes:
        process (process.TransportProcess): The transport process associated with the request.
        product (Optional[product.Product]): The product associated with the request.
        auxiliary (Optional[auxiliary.Auxiliary]): The auxiliary associated with the request.
        resource (Optional[resources.Resource]): The resource associated with the request to be the target of the transport of the auxiliaryand where it is needed.
    """

    def __init__(
        self,
        process: TransportProcess,
        product: Optional[Product]=None,
        auxiliary: Optional[Auxiliary]=None,
        resource: Optional[Resource]=None
    ):
        self.process: TransportProcess = process
        self.product: Optional[Product] = product
        self.auxiliary: Optional[Auxiliary] = auxiliary
        self.resource: Optional[Resource] = resource

        # TODO: rework this method. It is only used for interface homogenization -> restructure requests to be more general...
        self.origin: Optional[Locatable] = None
        self.target: Optional[Locatable] = None

    def set_route(self, route: List[Locatable]):
        """
        Caches a possible route of the transport request used later for setting the resource of the transport request.

        Args:
            process (process.TransportProcess): The process.
            route (List[product.Locatable]): The route.
        """
        # TODO: rework this method. It is only used for interface homogenization -> restructure requests to be more general...
        pass

class TransportResquest(Request):
    """
    Class to represents requests of a product for a transport process to be executed by a transport resource. Additionally, it contains the origin and target locations of the transport.

    Args:
        process (process.TransportProcess): The transport process.
        product (product.Product): The product.
        resource (resources.TransportResource): The transport resource.
        origin (product.Location): The origin location, either a resource, source or sink.
        target (product.Location): The target location, either a resource, source or sink.
    """
    def __init__(
        self,
        process: Union[TransportProcess, LinkTransportProcess],
        product: Union[Product, Auxiliary],
        resource: TransportResource,
        origin: Locatable,
        target: Locatable,
    ):
        self.process: Union[TransportProcess, LinkTransportProcess] = process
        self.product: Product = product
        self.resource: TransportResource = resource
        self.origin: Locatable = origin
        self.target: Locatable = target

        self.route: Optional[List[Locatable]] = None


    def set_process(self, process: PROCESS_UNION):
        """
        Sets the process of the request.

        Args:
            process (process.PROCESS_UNION): The process.
        """
        self.process = process
        # TODO: maybe do some special handling of compound processes here

    def copy_cached_routes(self, request: "TransportResquest"):
        """
        Copies the cached routes from another transport request.

        Args:
            request (TransportResquest): The transport request.
        """
        self.route = request.route

    def set_route(self, route: List[Locatable]):
        """
        Caches a possible route of the transport request used later for setting the resource of the transport request.

        Args:
            process (process.TransportProcess): The process.
            route (List[product.Locatable]): The route.
        """
        self.route = route

    def get_process(self) -> Union[TransportProcess, LinkTransportProcess]:
        """
        Returns the transport process of the transport request.

        Returns:
            process.TransportProcess: The transport process.
        """
        return self.process

    def get_resource(self) -> TransportResource:
        """
        Returns the transport resource of the transport request.

        Returns:
            resources.TransportResource: The transport resource.
        """
        return self.resource

    def get_origin(self) -> Locatable:
        """
        Returns the origin location of the transport request.

        Returns:
            product.Locatable: The origin location.
        """
        return self.origin

    def get_target(self) -> Locatable:
        """
        Returns the target location of the transport request.

        Returns:
            product.Locatable: The target location.
        """
        return self.target
    
    def get_route(self) -> List[Locatable]:
        """
        Returns the route of the transport request.

        Returns:
            List[product.Locatable]: The route.

        """
        return self.route



from __future__ import annotations
from typing import Generator, List
from pydantic import BaseModel


from simpy.resources import store

import logging
logger = logging.getLogger(__name__)

from prodsys.models import queue_data

from prodsys.simulation import sim


class Queue(store.FilterStore):
    """
    Class for storing products in a queue. The queue is a filter store with a limited or unlimited capacity, where product can be put and get from. 

    Args:
        env (simpy.Environment): The simulation environment.
        queue_data (queue_data.QueueData): The queue data object.

    Attributes:
        capacity (int, optional): The capacity of the queue. If 0 in the queue_data, the capacity is set to infinity.
        _pending_put (int): The number of products that are reserved for being put into the queue. Avoids bottleneck in the simulation.

    """
    def __init__(self, env: sim.Environment, data: queue_data.QueueData):
        self.env: sim.Environment = env
        self.data: queue_data.QueueData = data
        if data.capacity == 0:
            capacity = float("inf")
        else:
            capacity = data.capacity
        self._pending_put: int = 0
        super().__init__(env, capacity)
        self.state_change = self.env.event()


    def put(self, item) -> Generator:
        """
        Puts a product into the queue.

        Args:
            item (object): The product to be put into the queue.
        """
        self.unreseve()
        return_event = super().put(item)
        self.state_change.succeed()
        self.state_change = self.env.event()
        return return_event

    
    def get(self, filter) -> Generator:
        """
        Gets a product from the queue.

        Args:
            filter (Callable): The filter function to filter the items in the queue.

        Returns:
            object: The product that was gotten from the queue.
        """
        item = super().get(filter=filter)
        self.state_change.succeed()
        self.state_change = self.env.event()
        return item

    @property
    def full(self) -> bool:
        """
        Checks if the queue is full.

        Returns:
            bool: True if the queue is full, False otherwise.
        """
        logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "event": f"queue has {len(self.items)} items and {self._pending_put} pending puts for capacity {self.capacity}"})
        return (self.capacity - self._pending_put - len(self.items)) <= 0
    
    def reserve(self) -> None:
        """
        Reserves a spot in the queue for a product to be put into.

        Raises:
            RuntimeError: If the queue is full.
        """
        self._pending_put += 1
        if self._pending_put + len(self.items) > self.capacity:
            raise RuntimeError("Queue is full")
    
    def unreseve(self) -> None:
        """
        Unreserves a spot in the queue for a product to be put into after the put is completed.
        """
        self._pending_put -= 1
    
    def get_location(self):
        return self.data.location


"""
A Node is the express object for a NodeData object in the models API. It represents a location in the production system layout, 
used for LinkTransportProcesses to specify crossroads or specific locations for transport.
"""
from typing import Optional
from uuid import uuid1


from pydantic import ConfigDict, Field, conlist
from pydantic.dataclasses import dataclass

from prodsys.express import core

from prodsys.models import node_data

@dataclass
class Node(core.ExpressObject):
        """
        Represents a node data object of a link, which is just a location in the production system layout. 

        Attributes:
                location (List[float]): Location of the node. It has to be a list of length 2.
        """
        location: conlist(float, min_length=2, max_length=2) # type: ignore
        ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

        model_config=ConfigDict(json_schema_extra= {
                "examples": [
                        {
                                "ID": "N1",
                                "description": "Node 1",
                                "location": [0.0, 0.0],
                        }
                ]
        })
                
        def to_model(self) -> node_data.NodeData:
             """
             Function returns a NodeData object from the Node object.
             """
             return node_data.NodeData(
                  ID = self.ID,
                  location=self.location,
                  description="",)


"""
The `state` module contains the `prodsys.express` classes to represent the states that resources 
can be in during a simulation, besides processing and standby.

The following states are possible:

- `Breakdown`: A state that makes a resource unavailable for a certain time.
- `ProcessBreakdown`: A state that makes a process unavailable for a certain time but other processes can still be performed.
- `Setup`: A state that represents the time needed to change the process of a resource.	    

"""
from __future__ import annotations

from typing import List, Optional, Union
from uuid import uuid1

from abc import ABC

from pydantic import Field
from pydantic.dataclasses import dataclass

from prodsys.express import core, time_model

from prodsys.models import state_data



@dataclass
class State(ABC):
    """
    Abstract base class to represents a state.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the state.
    """

    time_model: time_model.TIME_MODEL_UNION

@dataclass
class BreakDownState(State, core.ExpressObject):
    """
    Class that represents a breakdown state.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Breakdwon occurence time model of the state.
        repair_time_model (time_model.TIME_MODEL_UNION): Reapit time model of the state.    
        ID (str): ID of the state.

    Attributes:
        type (state_data.StateTypeEnum): Type of the state. Equals to state_data.StateTypeEnum.BreakDownState.

    Examples:
        Breakdown state with a function time model:
        ``` py
        import prodsys.express as psx
        breakdown_time_model = psx.FunctionTimeModel(
            distribution_function="normal",
            location=10.0,
            scale=5.0,
        )
        repair_time_model = psx.FunctionTimeModel(
            distribution_function="normal",
            location=10.0,
            scale=5.0,
        )
        psx.BreakDownState(
            time_model=breakdown_time_model,
            reapit_time_model=repair_time_model
        )
        ```
    """
    repair_time_model: time_model.TIME_MODEL_UNION
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: state_data.StateTypeEnum = Field(default=state_data.StateTypeEnum.BreakDownState, init=False)

    def to_model(self) -> state_data.BreakDownStateData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            state_data.BreakDownStateData: Data object of the express object.
        """
        return state_data.BreakDownStateData(
            ID=self.ID,
            description="",
            time_model_id=self.time_model.ID,
            type=self.type,
            repair_time_model_id=self.repair_time_model.ID
        )
    
@dataclass
class ProcessBreakdownState(State, core.ExpressObject):
    """
    Class that represents a process breakdown state.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Breakdwon occurence time model of the state.
        repair_time_model (time_model.TIME_MODEL_UNION): Reapit time model of the state.
        process (process.PROCESS_UNION): Process that is broken down due to this breakdown state.
        ID (str): ID of the state.

    Attributes:
        type (state_data.StateTypeEnum): Type of the state. Equals to state_data.StateTypeEnum.ProcessBreakDownState.

    Examples:
        Process breakdown state with a normally distributed repair and breakdwon time for an examplary process:
       
        ``` py
        import prodsys.express as psx
        breakdown_time_model = psx.FunctionTimeModel(
            distribution_function="normal",
            location=100.0,
            scale=23.0,
        )
        repair_time_model = psx.FunctionTimeModel(
            distribution_function="normal",
            location=10.0,
            scale=5.0,
        )
        process_time_model = psx.FunctionTimeModel(
            distribution_function="normal",
            location=10.0,
            scale=5.0,
        )
        process = psx.ProdutionProcess(
            time_model=process_time_model
        )
        psx.ProcessBreakDownState(
            time_model=breakdown_time_model,
            repair_time_model=repair_time_model,
            process=process
        )
        ```
    """
    repair_time_model: time_model.TIME_MODEL_UNION
    process: process.PROCESS_UNION
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: state_data.StateTypeEnum = Field(default=state_data.StateTypeEnum.ProcessBreakDownState, init=False)

    def to_model(self) -> state_data.ProcessBreakDownStateData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            state_data.ProcessBreakDownStateData: Data object of the express object.
        """
        return state_data.ProcessBreakDownStateData(
            ID=self.ID,
            description="",
            time_model_id=self.time_model.ID,
            type=self.type,
            repair_time_model_id=self.repair_time_model.ID,
            process_id=self.process.ID
        )

@dataclass
class SetupState(State, core.ExpressObject):
    """
    Class that represents a setup state.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the state.
        origin_setup (process.Process): Process that is the origin of the setup.
        target_setup (process.Process): Process that is the target of the setup.
        ID (str): ID of the state.

    Attributes:
        type (state_data.StateTypeEnum): Type of the state. Equals to state_data.StateTypeEnum.SetupState.

    Examples:
        Setup state with a function time model:
        ``` py
        import prodsys.express as psx
        setup_time_model = psx.FunctionTimeModel(
            distribution_function="normal",
            location=10.0,
            scale=5.0,
        )
        dummy_process_time_model = psx.FunctionTimeModel(
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        dummy_origin_process = psx.ProductionProcess(
            time_model=dummy_process_time_model
        )
        dummy_target_process = psx.ProductionProcess(
            time_model=dummy_process_time_model
        )
        psx.SetupState(
            time_model=setup_time_model,
            origin_setup=dummy_origin_process,
            target_setup=dummy_target_process
        )
        ```
    """
    origin_setup: process.PROCESS_UNION
    target_setup: process.PROCESS_UNION
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: state_data.StateTypeEnum = Field(default=state_data.StateTypeEnum.SetupState, init=False)

    def to_model(self) -> state_data.SetupStateData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            state_data.SetupStateData: Data object of the express object.
        """
        return state_data.SetupStateData(
            ID=self.ID,
            description="",
            time_model_id=self.time_model.ID,
            type=self.type,
            origin_setup=self.origin_setup.ID,
            target_setup=self.target_setup.ID
        )
    
@dataclass
class ChargingState(State, core.ExpressObject):
    """
    Class that represents a charging state.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the state.
        battery_time_model (time_model.TIME_MODEL_UNION): Time model of the battery.
        ID (str): ID of the state.
    """
    battery_time_model: time_model.TIME_MODEL_UNION
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: state_data.StateTypeEnum = Field(default=state_data.StateTypeEnum.ChargingState, init=False)

    def to_model(self) -> state_data.ChargingStateData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            state_data.ChargingStateData: Data object of the express object.
        """
        return state_data.ChargingStateData(
            ID=self.ID,
            description="",
            time_model_id=self.time_model.ID,
            type=self.type,
            battery_time_model_id=self.battery_time_model.ID
        )
    
STATE_UNION = Union[BreakDownState, ProcessBreakdownState, SetupState, ChargingState]
from prodsys.express import process


from __future__ import annotations

from hashlib import md5
from typing import TYPE_CHECKING, Union, List, Dict

from pydantic import model_validator, root_validator

from prodsys.models.core_asset import CoreAsset
from prodsys.models import source_data

if TYPE_CHECKING:
    from prodsys.adapters.adapter import ProductionSystemAdapter


class AuxiliaryData(CoreAsset):
    """
    Class that represents auxiliary components required to process or transport a product by a resource..

    Some examples are workpiece carriers or tools.

    Args:
        ID (str): ID of the auxiliary component.
        description (str): Description of the auxiliary component.
        router (str): Router of the auxiliary component.
        transport_process (str): Transport process of the auxiliary component.
        storage_queues (List[str], optional): List of storage queues where the auxiliary component is stored between usages. Defaults to [].
        quantity_in_storages (List[int], optional): List of initial quantities in the storage queues, in sequence of the storage queues. Defaults to [].
        relevant_processes (List[str], optional): List of relevant processes where the auxiliary component is needed. Defaults to [], meaning all processes.
        relevant_transport_processes (List[str], optional): List of relevant transport processes where the auxiliary component is needed. Defaults to [], meaning all processes.
    """
    auxiliary_type: str
    transport_process: str
    storages: List[str]
    quantity_in_storages: List[int]
    relevant_processes: List[str] = []
    relevant_transport_processes: List[str] = []

    @model_validator(mode="before")
    def check_processes(cls, values):
        if "auxiliary_type" in values and values["auxiliary_type"]:
            values["ID"] = values["auxiliary_type"]
        else:
            values["auxiliary_type"] = values["ID"]
        return values

    def hash(self, adapter: ProductionSystemAdapter) -> str:
        """
        Function to hash the auxiliary component.

        Returns:
            str: Hash of the auxiliary component.
        """
        transport_processes_hash = ""
        storages_hashes = []
        relevant_processes_hashes = []
        relevant_transport_processes_hashes = []

        for queue in adapter.queue_data:
            if queue.ID in self.storages:
                storages_hashes.append(queue.hash())
        for process in adapter.process_data:
            if process.ID in self.relevant_processes:
                relevant_processes_hashes.append(process.hash(adapter))
            if process.ID in self.relevant_transport_processes:
                relevant_transport_processes_hashes.append(process.hash(adapter))
            if process.ID == self.transport_process:
                transport_processes_hash = process.hash(adapter)
        
        return md5("".join([*map(str, [self.auxiliary_type, self.quantity_in_storages, transport_processes_hash, *storages_hashes, *self.quantity_in_storages, *relevant_processes_hashes, *relevant_transport_processes_hashes])]).encode("utf-8")).hexdigest()


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Union, List, Optional

from pydantic import BaseModel



if TYPE_CHECKING:
    from prodsys.simulation.resources import ProductionResource, TransportResource
    from prodsys.simulation.source import Source
    from prodsys.simulation.sink import Sink
    from prodsys.simulation.node import Node
    from prodsys.simulation import request
from prodsys.simulation import route_finder, time_model

from prodsys.models import processes_data


class Process(ABC, BaseModel):
    """
    Abstract process base class that defines the interface for all processes.
    """

    process_data: processes_data.PROCESS_DATA_UNION
    time_model: Optional[time_model.TimeModel]

    @abstractmethod
    def matches_request(self, request: request.Request) -> bool:
        """
        Returns True if the process matches the request.

        Args:
            request (request.Request): The request.

        Returns:
            bool: True if the process matches the request.
        """
        pass

    @abstractmethod
    def get_process_time(self, *args) -> float:
        """
        Returns the time it takes to execute the process.

        Returns:
            float: Time it takes to execute the process.
        """
        pass

    @abstractmethod
    def get_expected_process_time(self, *args) -> float:
        """
        Returns the expected time it takes to execute the process.

        Returns:
            float: Expected time it takes to execute the process.
        """
        pass


class ProductionProcess(Process):
    """Class that represents a production process.

    Args:
        process_data (processes_data.ProductionProcessData): The process data.
        time_model (time_model.TimeModel): The time model.
    """

    process_data: processes_data.ProductionProcessData

    def matches_request(self, request: request.Request) -> bool:
        requested_process = request.process
        if not isinstance(requested_process, ProductionProcess) and not isinstance(
            requested_process, CompoundProcess
        ):
            return False
        if isinstance(requested_process, CompoundProcess):
            return self.process_data.ID in requested_process.process_data.process_ids
        return requested_process.process_data.ID == self.process_data.ID

    def get_process_time(self) -> float:
        return self.time_model.get_next_time()

    def get_expected_process_time(self) -> float:
        return self.time_model.get_expected_time()


class CapabilityProcess(Process):
    """
    Class that represents a capability process.

    Args:
        process_data (processes_data.CapabilityProcessData): The process data.
        time_model (time_model.TimeModel): The time model.
    """
    process_data: processes_data.CapabilityProcessData

    def matches_request(self, request: request.Request) -> bool:
        requested_process = request.process
        if (
            not is_process_with_capability(requested_process) and not isinstance(requested_process, CompoundProcess)
        ):
            return False
        if isinstance(requested_process, CompoundProcess):
            return self.process_data.capability in [
                p.process_data.capability
                for p in requested_process.contained_processes_data
                if isinstance(p, CapabilityProcess)
            ]
        return requested_process.process_data.capability == self.process_data.capability

    def get_process_time(self) -> float:
        return self.time_model.get_next_time()

    def get_expected_process_time(self) -> float:
        return self.time_model.get_expected_time()


class TransportProcess(Process):
    """
    Class that represents a transport process.

    Args:
        process_data (processes_data.TransportProcessData): The process data.
        time_model (time_model.TimeModel): The time model.
    """
    process_data: processes_data.TransportProcessData

    def matches_request(self, request: request.TransportResquest) -> bool:
        requested_process = request.process
        if not isinstance(requested_process, TransportProcess) and not isinstance(
            requested_process, CompoundProcess
        ):
            return False
        if isinstance(requested_process, TransportProcess) and not requested_process.process_data.ID == self.process_data.ID:
            return False
        if isinstance(requested_process, CompoundProcess) and not self.process_data.ID in requested_process.process_data.process_ids:
            return False
        request.set_route(route=[request.origin, request.target])
        return True

    def get_process_time(self, origin: List[float], target: List[float]) -> float:
        return self.time_model.get_next_time(origin=origin, target=target)

    def get_expected_process_time(self, *args) -> float:
        return self.time_model.get_expected_time(*args)
    
def is_process_with_capability(process: "PROCESS_UNION") -> bool:
    """
    Returns True if the given process is a process with capability.

    Args:
        process (PROCESS_UNION): The process.

    Returns:
        bool: True if the given process is a process with capability.
    """
    return isinstance(process, CapabilityProcess) or isinstance(process, RequiredCapabilityProcess) or (isinstance(process, LinkTransportProcess) and process.process_data.capability)

def is_available_process_with_capability(process: "PROCESS_UNION") -> bool:
    """
    Returns True if the given process is an available process with capability.

    Args:
        process (PROCESS_UNION): The process.

    Returns:
        bool: True if the given process is an available process with capability.
    """
    return isinstance(process, CapabilityProcess) or (isinstance(process, LinkTransportProcess) and process.process_data.capability)

class CompoundProcess(Process):
    """
    Class that represents a compound process.

    Args:
        process_data (processes_data.CompoundProcessData): The process data.
        processes (List[Union[ProductionProcess, TransportProcess, CapabilityProcess, RequiredCapabilityProcess]]): The processes.
    """
    process_data: processes_data.CompoundProcessData
    contained_processes_data: List[
        Union[
            processes_data.ProductionProcessData,
            processes_data.TransportProcessData,
            processes_data.CapabilityProcessData,
            processes_data.RequiredCapabilityProcessData,
            processes_data.LinkTransportProcessData,
        ]
    ]

    def matches_request(self, request: request.Request) -> bool:
        requested_process = request.process
        if isinstance(requested_process, ProductionProcess) or isinstance(
            requested_process, TransportProcess
        ) or (isinstance(requested_process, LinkTransportProcess) and not requested_process.process_data.capability):
            return requested_process.process_data.ID in self.process_data.process_ids
        elif is_process_with_capability(requested_process):
            return requested_process.process_data.capability in [
                p.process_data.capability
                for p in self.contained_processes_data
                if is_available_process_with_capability(p)
            ]
        elif isinstance(requested_process, CompoundProcess):
            return any(
                p.ID in self.process_data.process_ids
                for p in requested_process.contained_processes_data
                if isinstance(p, ProductionProcess)
            ) or any(
                p.capability
                in [
                    p.process_data.capability
                    for p in self.contained_processes_data
                    if is_available_process_with_capability(p)
                ]
                for p in requested_process.contained_processes_data
                if is_available_process_with_capability(p)
            )
        return False

    def get_process_time(self) -> float:
        raise NotImplementedError("CompoundProcess does not have a process time.")

    def get_expected_process_time(self) -> float:
        raise NotImplementedError("CompoundProcess does not have a process time.")


class RequiredCapabilityProcess(Process):
    """
    Class that represents a required capability process.

    Args:
        process_data (processes_data.RequiredCapabilityProcessData): The process data.
    """
    process_data: processes_data.RequiredCapabilityProcessData

    def matches_request(self, request: request.Request) -> bool:
        raise NotImplementedError("RequiredCapabilityProcess does not match requests but only generates them.")

    def get_process_time(self) -> float:
        raise NotImplementedError(
            "RequiredCapabilityProcess does not have a process time."
        )

    def get_expected_process_time(self) -> float:
        raise NotImplementedError(
            "RequiredCapabilityProcess does not have a process time."
        )

class LinkTransportProcess(TransportProcess):
    """
    Class that represents a transport link process.
    """
    process_data: processes_data.LinkTransportProcessData
    links: Optional[List[List[Union[Node, Source, Sink, ProductionResource]]]]

    def matches_request(self, request: request.TransportResquest) -> bool:
        requested_process = request.process

        if not isinstance(requested_process, LinkTransportProcess) and not isinstance(requested_process, RequiredCapabilityProcess) and not isinstance(
            requested_process, CompoundProcess
        ):
            return False
        
        if isinstance(requested_process, CompoundProcess):
            possible_processes = []
            for process_data_instance in requested_process.contained_processes_data:
                if is_process_with_capability(process_data_instance) and self.process_data.capability and self.process_data.capability == process_data_instance.capability:
                    possible_processes.append(process_data_instance)
                if isinstance(process_data_instance, LinkTransportProcess) and process_data_instance.ID == self.process_data.ID:
                    possible_processes.append(process_data_instance)
            if not possible_processes:
                return False

        if is_process_with_capability(requested_process):
            if not requested_process.process_data.capability == self.process_data.capability:
                return False
        
        if isinstance(requested_process, LinkTransportProcess):
            if not requested_process.process_data.ID == self.process_data.ID:
                return False
            
        route = route_finder.find_route(request=request, process=self)
        if not route:
            return False
        return True
    
    def get_process_time(self, request: request.TransportResquest) -> float:
        route = request.get_route()
        total_time = 0
        link_route = [route[i:i+2] for i in range(len(route)-1)]

        for link in link_route:
            time = self.time_model.get_next_time(origin=link[0].get_location(), target=link[1].get_location())
            total_time += time
        return total_time

    def get_expected_process_time(self, *args) -> float:
        return self.time_model.get_expected_time(*args)
    
    

PROCESS_UNION = Union[
    CompoundProcess,
    RequiredCapabilityProcess,
    ProductionProcess,
    TransportProcess,
    CapabilityProcess,
    LinkTransportProcess,
]
"""
Union type for all processes.
"""
from prodsys.simulation.resources import ProductionResource, TransportResource
from prodsys.simulation.source import Source
from prodsys.simulation.sink import Sink
from prodsys.simulation.node import Node
from prodsys.simulation import request
# LinkTransportProcess.model_rebuild()



"""
The `time_model` module contains classes to specify time models in the simulation for the arrival
of products, performance of processes and transports and the duration of states.

The following time models are possible:
- `SequentialTimeModelData`: A time model that is based on a sequence of values.
- `FunctionTimeModelData`: A time model that is based on a distribution function which gets sampled.
- `ManhattanDistanceTimeModelData`: A time model that is based on the manhattan distance between two nodes and a constant velocity.
"""
# from __future__ import annotations

from hashlib import md5
from typing import List, Literal, Union
from typing_extensions import deprecated
from enum import Enum

from pydantic import ConfigDict, Field

from prodsys.models.core_asset import CoreAsset
from prodsys.util.statistical_functions import FunctionTimeModelEnum


class TimeModelEnum(str, Enum):
    """
    Enum that represents the different kind time models.

    - FunctionTimeModel: A time model that is based on a distribution function which gets sampled.
    - SampleTimeModel: A time model that samples values from a provided data set by random choice of a sample element.
    - ScheduledTimeModel: A time model that is based on a schedule of timely values. Should only be used for arrival time models of sources.
    - DistanceTimeModel: A time model that is based on the distance between two nodes and a constant velocity and a distance metric.
    - SequentialTimeModel: A time model that is based on a sequence of values. Deprecated, use SampleTimeModel instead.
    - ManhattanDistanceTimeModel: A time model that is based on the manhattan distance between two nodes and a constant velocity. Deprecated, use DistanceTimeModel instead.
    """

    FunctionTimeModel = "FunctionTimeModel"
    SampleTimeModel = "SampleTimeModel"
    ScheduledTimeModel = "ScheduledTimeModel"
    DistanceTimeModel = "DistanceTimeModel"
    # TODO: remove in the future
    SequentialTimeModel = "SequentialTimeModel"
    ManhattanDistanceTimeModel = "ManhattanDistanceTimeModel"




@deprecated(
    'prodsys SequentialTimeModelData is deprecated and will be removed in the future. '
    'Use prodsys.models.time_model_data.SampleTimeModel instead.',
    category=None,
)
class SequentialTimeModelData(CoreAsset):
    """
    Class that represents a time model that is based on a sequence of values.

    Args:
        ID (str): ID of the time model.
        description (str): Description of the time model.
        sequence (List[float]): Sequence of time values.

    Examples:
        Sequential time model with 7 time values:
        ```	py
        import prodsys
        prodsys.time_model_data.SequentialTimeModelData(
            ID="sequence_time_model_1",
            description="Examplary sequence time model",
            sequence=[25.0, 13.0, 15.0, 16.0, 17.0, 20.0, 21.0],
        )
        ```
    """

    sequence: List[float]

    def hash(self) -> str:
        """
        Returns a unique hash for the time model considering its sequence. Can be used to compare time models for equal functionality.

        Returns:
            str: Hash of the time model.
        """
        return md5(("".join([*map(str, self.sequence)])).encode("utf-8")).hexdigest()
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "sequence_time_model_1",
                "description": "Examplary sequence time model",
                "sequence": [25.0, 13.0, 15.0, 16.0, 17.0, 20.0, 21.0],
            },
        ]
    
    })


class SampleTimeModelData(CoreAsset):
    """
    Class that represents a time model that samples values from a provided data set by random choice of a sample element.

    Args:
        ID (str): ID of the time model.
        description (str): Description of the time model.
        samples (List[float]): List of sample time values.

    Examples:
        Sample time model with 7 time values:
        ```	py
        import prodsys
        prodsys.time_model_data.SampleTimeModelData(
            ID="sample_time_model_1",
            description="Examplary sample time model",
            samples=[25.0, 13.0, 15.0, 16.0, 17.0, 20.0, 21.0],
        )
        ```
    """

    samples: List[float]

    def hash(self) -> str:
        """
        Returns a unique hash for the time model considering its samples. Can be used to compare time models for equal functionality.

        Returns:
            str: Hash of the time model.
        """
        return md5(("".join([*map(str, self.samples)])).encode("utf-8")).hexdigest()
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "sample_time_model_1",
                "description": "Examplary sample time model",
                "samples": [25.0, 13.0, 15.0, 16.0, 17.0, 20.0, 21.0],
            },
        ]
    
    })


class ScheduledTimeModelData(CoreAsset):
    """
    Class that represents a time model that is based on a schedule of timely values. Should only be used for arrival time models of sources.

    Args:
        ID (str): ID of the time model.
        description (str): Description of the time model.
        schedule (List[float]): Schedule of time values.
        absolute (bool): If the schedule contains absolute time values in simulation time or relative time differences. Default is False.
        cyclic (bool): If the schedule should be repeated cyclically. If False, the time model does not return any values after the schedule is exhausted. Default is False.

    Examples:
        Scheduled time model with 7 time values:
        ```	py
        import prodsys
        prodsys.time_model_data.ScheduledTimeModelData(
            ID="scheduled_time_model_1",
            description="Examplary scheduled time model",
            schedule=[3.0, 5.0, 7.0, 9.0, 11.0, 13.0, 15.0],
            absolute=True,
            cyclic=False
        )
        ```
    """

    schedule: List[float]
    absolute: bool
    cyclic: bool = False

    def hash(self) -> str:
        """
        Returns a unique hash for the time model considering its schedule and boolean values. Can be used to compare time models for equal functionality.

        Returns:
            str: Hash of the time model.
        """
        return md5(("".join([*map(str, self.schedule)] + [*map(str, [self.absolute, self.cyclic])])).encode("utf-8")).hexdigest()
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "scheduled_time_model_1",
                "description": "Examplary scheduled time model",
                "schedule": [3.0, 5.0, 7.0, 9.0, 11.0, 13.0, 15.0],
                "absolute": True,
                "cyclic": False
            },
        ]
    
    })


class FunctionTimeModelData(CoreAsset):
    """
    Class that represents a time model that is based on a function and represents the timely values by their distribution function.

    Args:
        ID (str): ID of the time model.
        description (str): Description of the time model.
        distribution_function (FunctionTimeModelEnum): Zype of the distribution function of the time model.
        location (float): Location parameter of the distribution function.
        scale (float): Scale parameter of the distribution function.

    Examples:
        Normal distribution time model with 20 minutes:
        ``` py
        import prodsys
        prodsys.time_model_data.FunctionTimeModelData(
            ID="function_time_model_1",
            description="normal distribution time model with 20 minutes",
            distribution_function=prodsys.FunctionTimeModelEnum.normal,
            location=20.0,
            scale=5.0,
        )
        ```
    """

    distribution_function: FunctionTimeModelEnum
    location: float
    scale: float
    batch_size: int = Field(default=100, init=False)

    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "function_time_model_1",
                "description": "normal distribution time model with 20 minutes",
                "distribution_function": "normal",
                "location": 20.0,
                "scale": 5.0,
            },
        ]
    
    })
    
    def hash(self) -> str:
        """
        Returns a unique hash for the time model considering its distribution function, location and scale. Can be used to compare time models for equal functionality.

        Returns:
            str: Hash of the time model.
        """
        return md5(("".join([*map(str, [self.distribution_function, self.location, self.scale])])).encode("utf-8")).hexdigest()
    
@deprecated(
    'prodsys ManhattanDistanceTimeModelData is deprecated and will be removed in the future. Use instead the prodsys.models.time_model_data.DistanceTimeModelData.',
    category=None,
)
class ManhattanDistanceTimeModelData(CoreAsset):
    """
    Class that represents a time model that is based on the manhattan distance between two nodes and a constant velocity.

    Args:
        ID (str): ID of the time model.
        description (str): Description of the time model.
        speed (float): Speed of the transport.
        reaction_time (float): Reaction time of the transport.

    Examples:
        Manhattan time model with speed 180 m/min = 3 m/s and reaction time 0.15 minutes:
        ``` py
        import prodsys
        time_model_data.ManhattanDistanceTimeModelData(
            ID="manhattan_time_model_1",
            description="manhattan time model with speed 180 m/min = 3 m/s",
            speed=180.0,
            reaction_time=0.15,
        )
    """

    speed: float
    reaction_time: float

    def hash(self) -> str:
        """
        Returns a unique hash for the time model considering its speed and reaction time. Can be used to compare time models for equal functionality.

        Returns:
            str: Hash of the time model.
        """
        return md5(("".join([*map(str, [self.speed, self.reaction_time])])).encode("utf-8")).hexdigest()
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "manhattan_time_model_1",
                "description": "manhattan time model with speed 180 m/min = 3 m/s",
                "speed": 180.0,
                "reaction_time": 0.15,
            },
        ]
    })

class DistanceTimeModelData(CoreAsset):
    """
    Class that represents a time model that is based on the distance between two nodes and a constant velocity.
    The distance is calculated based on the metric provided.

    Args:
        ID (str): ID of the time model.
        description (str): Description of the time model.
        speed (float): Speed of the transport.
        reaction_time (float): Reaction time of the transport.
        metric (Literal["manhattan", "euclidian"]): Metric to calculate the distance. Default is "manhattan".

    Examples:
        Distance time model with speed 180 m/min = 3 m/s and reaction time 0.15 minutes:
        ``` py
        import prodsys
        time_model_data.DistanceTimeModelData(
            ID="distance_time_model_1",
            description="distance time model with speed 180 m/min = 3 m/s",
            speed=180.0,
            reaction_time=0.15,
            metric="manhattan",
        )
    """

    speed: float
    reaction_time: float
    metric: Literal["manhattan", "euclidian"] = "manhattan"

    def hash(self) -> str:
        """
        Returns a unique hash for the time model considering its speed and reaction time. Can be used to compare time models for equal functionality.

        Returns:
            str: Hash of the time model.
        """
        return md5(("".join([*map(str, [self.speed, self.reaction_time, self.metric])])).encode("utf-8")).hexdigest()
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "distance_time_model_1",
                "description": "distance time model with speed 180 m/min = 3 m/s",
                "speed": 30.0,
                "reaction_time": 0.15,
                "metric": "manhattan",
            },
        ]
    })


TIME_MODEL_DATA = Union[
    FunctionTimeModelData, SampleTimeModelData, ScheduledTimeModelData, DistanceTimeModelData, SequentialTimeModelData, ManhattanDistanceTimeModelData, 
]



"""
The `time_model` module contains classes to specify time models in the simulation for the arrival
of products, performance of processes and the duration of states.

The following time models are possible:
- `SequentialTimeModel`: A time model that is based on a sequence of values.
- `FunctionTimeModel`: A time model that is based on a distribution function which gets sampled.
- `ManhattanDistanceTimeModel`: A time model that is based on the manhattan distance between two nodes and a constant velocity.
"""
from typing import List, Literal, Optional, Union
from typing_extensions import deprecated
from uuid import uuid1

from pydantic.dataclasses import dataclass
from pydantic import Field

from prodsys.express import core

from prodsys.models import time_model_data

@deprecated(
    "The SequentialTimeModel is deprecated and will be removed in the next version. Please use the SampleTimeModel instead.",
    category=None
)
@dataclass
class SequentialTimeModel(core.ExpressObject):
    """
    Class that represents a time model that is based on a sequence of values.

    Args:
        sequence (List[float]): Sequence of time values.
        ID (str): ID of the time model.

    Examples:
        Sequential time model with 7 time values:
        ``` py
        import prodsys.express as psx
        psx.SequentialTimeModel(
            sequence=[25.0, 13.0, 15.0, 16.0, 17.0, 20.0, 21.0],
        )
        ```
    """
    sequence: List[float]
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))


    def to_model(self) -> time_model_data.SequentialTimeModelData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            time_model_data.SequentialTimeModelData: Data object of the express object.
        """
        return time_model_data.SequentialTimeModelData(
            sequence=self.sequence,
            ID=self.ID,
            description=""
        )
    
@dataclass
class SampleTimeModel(core.ExpressObject):
    """
    Class that represents a time model that is based on a list of sample values.

    Args:
        samples (List[float]): Values to sample from.
        ID (str): ID of the time model.

    Examples:
        Sample time model with 7 time values:
        ``` py
        import prodsys.express as psx
        psx.SampleTimeModel(
            samples=[25.0, 13.0, 15.0, 16.0, 17.0, 20.0, 21.0],
        )
        ```
    """
    samples: List[float]
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))


    def to_model(self) -> time_model_data.SampleTimeModelData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            time_model_data.SequentialTimeModelData: Data object of the express object.
        """
        return time_model_data.SampleTimeModelData(
            samples=self.samples,
            ID=self.ID,
            description=""
        )
    
@dataclass
class ScheduledTimeModel(core.ExpressObject):
    """
    Class that represents a time model that is based on a schedule of timely values.
    If the schedule is not cyclic, usage in other classes than sources can lead to unexpected behavior.

    Args:
        schedule (List[float]): Schedule of time values.
        absolute (bool): If the schedule contains absolute time values or relative values to each other.
        cyclic (bool): If the schedule is cyclic or not.
        ID (str): ID of the time model.

    Examples:
        Scheduled time model with 7 time values:
        ``` py
        import prodsys.express as psx
        psx.ScheduledTimeModel(
            schedule=[25.0, 13.0, 15.0, 16.0, 17.0, 20.0, 21.0],
        )
        ```
    """
    schedule: List[float]
    absolute: bool = True
    cyclic: bool = True
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))


    def to_model(self) -> time_model_data.ScheduledTimeModelData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            time_model_data.SequentialTimeModelData: Data object of the express object.
        """
        return time_model_data.ScheduledTimeModelData(
            schedule=self.schedule,
            absolute=self.absolute,
            cyclic=self.cyclic,
            ID=self.ID,
            description=""
        )

@dataclass
class FunctionTimeModel(core.ExpressObject):
    """
    Class that represents a time model that is based on a function and represents the timely values by their distribution function.

    Args:
        distribution_function (FunctionTimeModelEnum): Function that represents the time model, can either be 'normal', 'lognormal', 'exponential' or 'constant'.
        location (float): Location parameter of the distribution function.
        scale (float): Scale parameter of the distribution function.
        ID (str): ID of the time model.


    Examples:
        Normal distribution time model with 20 minutes mean and 5 minutes standard deviation:
        ``` py
        import prodsys.express as psx
        psx.FunctionTimeModel(
            distribution_function="normal",
            location=20.0,
            scale=5.0,
        )
        ```
    """
    distribution_function: time_model_data.FunctionTimeModelEnum
    location: float
    scale: float = 0.0
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    def to_model(self) -> time_model_data.FunctionTimeModelData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            time_model_data.FunctionTimeModelData: Data object of the express object.
        """
        return time_model_data.FunctionTimeModelData(
            distribution_function=self.distribution_function,
            location=self.location,
            scale=self.scale,
            ID=self.ID,
            description=""
        )
    

@dataclass
class DistanceTimeModel(core.ExpressObject):
    """
    Class that represents a time model that is based on the distance between two nodes and a constant velocity.
    The distance is calculated by the specified distance metric.

    Args:
        speed (float): Speed of the vehicle in meters per minute.
        reaction_time (float): Reaction time of the driver in minutes.
        metric (Literal["manhattan", "euclidean"]): Metric to calculate the distance.

    Examples:
        Distance time model with a speed of 50 meters per minute and a reaction time of 0.5 minutes, using the default manhattan distance metric:
        ``` py
        import prodsys.express as psx
        psx.DistanceTimeModel(
            speed=50.0,
            reaction_time=0.5,
        )
        ```
    """
    speed: float
    reaction_time: float
    metric: Literal["manhattan", "euclidean"] = "manhattan"
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    def to_model(self) -> time_model_data.DistanceTimeModelData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            time_model_data.DistanceTimeModelData: Data object of the express object.
        """
        return time_model_data.DistanceTimeModelData(
            speed=self.speed,
            reaction_time=self.reaction_time,
            metric=self.metric,
            ID=self.ID,
            description=""
        )

@deprecated(
    "The ManhattanDistanceTimeModel is deprecated and will be removed in the next version. Please use the DistanceTimeModel instead.",
    category=None
)
@dataclass
class ManhattanDistanceTimeModel(core.ExpressObject):
    """
    Class that represents a time model that is based on the manhattan distance between two nodes and a constant velocity.

    Args:
        speed (float): Speed of the vehicle in meters per minute.
        reaction_time (float): Reaction time of the driver in minutes.

    Examples:
        Manhattan distance time model with a speed of 50 meters per minute and a reaction time of 0.5 minutes:
        ``` py
        import prodsys.express as psx
        psx.ManhattenDistanceTimeModel(
            speed=50.0,
            reaction_time=0.5,
        )
        ```
    """
    speed: float
    reaction_time: float
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    def to_model(self) -> time_model_data.ManhattanDistanceTimeModelData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            time_model_data.ManhattanDistanceTimeModelData: Data object of the express object.
        """
        return time_model_data.ManhattanDistanceTimeModelData(
            speed=self.speed,
            reaction_time=self.reaction_time,
            ID=self.ID,
            description=""
        )

TIME_MODEL_UNION = Union[
    SampleTimeModel,
    ScheduledTimeModel,
    DistanceTimeModel,
    FunctionTimeModel,
    SequentialTimeModel,
    ManhattanDistanceTimeModel,
]
    






from __future__ import annotations

from typing import Optional, TYPE_CHECKING, Union


from prodsys.factories import process_factory, source_factory, sink_factory, resource_factory, node_factory
from prodsys.simulation import process

if TYPE_CHECKING:
    from prodsys.simulation import resources, sink, source, node


class LinkTransportProcessUpdater:
    """
    Updater class that updates the links of `prodsys.simulation` LinkTransportProcess objects in the process factory based on the created node, resource, source and sink objects.

    Args:
        process_factory (process_factory.ProcessFactory): Factory that contains alll process objects.
        source_factory (source_factory.SourceFactory): Factory that contains all source objects.
        sink_factory (sink_factory.SinkFactory): Factory that contains all sink objects.
        resource_factory (resource_factory.ResourceFactory): Factory that contains all resource objects.
        node_factory (node_factory.NodeFactory): Factory that contains all node objects.
    """
    def __init__(self, process_factory: process_factory.ProcessFactory, source_factory: source_factory.SourceFactory, sink_factory: sink_factory.SinkFactory, resource_factory: resource_factory.ResourceFactory, node_factory: node_factory.NodeFactory):
        self.process_factory = process_factory
        self.source_factory = source_factory
        self.sink_factory = sink_factory
        self.resource_factory = resource_factory
        self.node_factory = node_factory

    def update_links_with_objects(self):
        """
        The method updates the links of the LinkTransportProcess objects in the process factory with the created node, resource, source and sink objects from the respective factories.
        """
        for process_instance in self.process_factory.processes:
            if isinstance(process_instance, process.LinkTransportProcess):
                self.update_links(process_instance)

    def update_links(self, process_instance: process.LinkTransportProcess):
        link_id_list = process_instance.process_data.links
        links_list = []

        for link in link_id_list:
            start = link[0]
            end = link[1]

            start_obj = self.get_node_resource_source_sink(start)
            if not start_obj:
                raise ValueError(f"LinkTransportProcessUpdater: Could not find object with ID {start} for a link in Process {process_instance.process_data.ID}.")
            end_obj = self.get_node_resource_source_sink(end)
            if not end_obj:
                raise ValueError(f"LinkTransportProcessUpdater: Could not find object with ID {end} for a link in Process {process_instance.process_data.ID}.")

            links_list.append([start_obj, end_obj])

        process_instance.links = links_list

    def get_node_resource_source_sink(self, ID: str) -> Optional[Union[node.Node, resources.Resource, source.Source, sink.Sink]]:
        try:
            node = self.node_factory.get_node(ID)
            return node
        except IndexError:
            pass
        try:
            resource = self.resource_factory.get_resource(ID)
            return resource
        except IndexError:
            pass
        try:
            source = self.source_factory.get_source(ID)
            return source
        except IndexError:
            pass
        try:
            sink = self.sink_factory.get_sink(ID)
            return sink
        except IndexError:
            pass



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List, Optional, Dict
import random
from pydantic import BaseModel, Field

from prodsys.util.util import flatten

if TYPE_CHECKING:
    from prodsys.simulation.process import PROCESS_UNION


class ProcessModel(ABC, BaseModel):
    """
    Abstract process model base class that defines the interface for all process models.
    """

    @abstractmethod
    def get_next_possible_processes(self) -> Optional[List[PROCESS_UNION]]:
        """
        Returns the next possible processes.

        Returns:
            Optional[List[PROCESS_UNION]]: List of possible processes.
        """
        pass

    @abstractmethod
    def update_marking_from_transition(
        self, chosen_process: PROCESS_UNION
    ) -> None:
        """
        Updates the marking of the process model based on the chosen process.

        Args:
            chosen_process (PROCESS_UNION): The chosen process that is executed.
        """
        pass


class ListProcessModel(ProcessModel):
    """
    Process model that is based on a list of processes. The processes are executed sequentially in the order of the list.

    Args:
        process_list (List[PROCESS_UNION]): List of processes that are executed sequentially.
    """

    process_list: List[PROCESS_UNION]
    current_marking: int = Field(default=0, init=False)

    def get_next_possible_processes(self) -> Optional[List[PROCESS_UNION]]:
        if self.current_marking == len(self.process_list):
            return None
        return [self.process_list[self.current_marking]]

    def update_marking_from_transition(
        self, chosen_process: PROCESS_UNION
    ) -> None:
        self.current_marking += 1

class PrecendeGraphNode(BaseModel):
    """
    Class that represents a node in a precedence graph.

    Args:
        process (PROCESS_UNION): The process that is represented by the node.
        successors (Optional[List[PrecendeGraphNode]]): List of successor nodes.
        predecessors (Optional[List[PrecendeGraphNode]]): List of predecessor nodes.

    Attributes:
        marking (bool): Indicates if the node is marked.
    """

    process: PROCESS_UNION
    successors: Optional[List[PrecendeGraphNode]] = []
    predecessors: Optional[List[PrecendeGraphNode]] = []
    marking: bool = Field(default=False, init=False)

    def update_marking(self):
        """
        Updates the marking of the node, to save that the process has been executed.
        """
        self.marking = True


def get_predecessor_processes(
    target_process_id: str, adjacency_matrix: Dict[str, List[str]]
) -> List[PROCESS_UNION]:
    """
    Returns the predecessing processes' IDs of a ID of a process.

    Args:
        target_process_id (str): process ID of the process for which the predecessing processes' IDs are returned.
        adjacency_matrix (Dict[str, List[str]]): Adjacency matrix of the process model. The keys are the process IDs and the values are the IDs of the successor processes.

    Returns:
        List[PROCESS_UNION]: List of predecessing processes' IDs.
    """
    predecessors = []
    for process_id, successors in adjacency_matrix.items():
        if target_process_id in successors:
            predecessors.append(process_id)
    return predecessors


def get_predecessors_adjacency_matrix(
    adjacency_matrix: Dict[str, List[str]]
) -> Dict[str, List[str]]:
    """
    Returns the predecessing processes' IDs of all processes in a process model.

    Args:
        adjacency_matrix (Dict[str, List[str]]): Adjacency matrix of the process model. The keys are the process IDs and the values are the IDs of the successor processes.

    Returns:
        Dict[str, List[str]]: Predecessor adjacency matrix of the process model. The keys are the process IDs and the values are the IDs of the predecessing processes.
    """
    predecessors_adjacency_matrix = {}
    for process_id in adjacency_matrix.keys():
        predecessors_adjacency_matrix[process_id] = get_predecessor_processes(
            process_id, adjacency_matrix
        )
    return predecessors_adjacency_matrix


def get_adjacency_matrix_from_edges(edges: List[List[str]]) -> Dict[str, List[str]]:
    """
    Returns the adjacency matrix of a process model from a list of edges.

    Args:
        edges (List[List[str]]): List of edges. Each edge is a list of two process IDs. The first process ID is the predecessor and the second process ID is the successor.

    Returns:
        Dict[str, List[str]]: Adjacency matrix of the process model. The keys are the process IDs and the values are the IDs of the successor processes.
    """
    nodes = list(set(flatten(edges)))
    adjacency_matrix = {}
    for node in nodes:
        successors = [edge[1] for edge in edges if edge[0] == node]
        adjacency_matrix[node] = successors
    return adjacency_matrix


class PrecedenceGraphProcessModel(ProcessModel):
    """
    Process model that is based on a precedence graph.

    Attributes:
        nodes (List[PrecendeGraphNode]): List of nodes in the precedence graph.
        current_marking (Optional[PrecendeGraphNode]): The current marking, i.e. the node that represents the previously executed process, of the process model.
    """

    nodes: List[PrecendeGraphNode] = Field(default_factory=list, init=False)
    current_marking: Optional[PrecendeGraphNode] = Field(init=False, default=None)

    def __str__(self) -> str:
        """
        Returns a string representation of the adjacency matrix of the process model.

        Returns:
            str: String representation of the adjacency matrix of the process model.
        """
        adjacency_matrix = {}
        for node in self.nodes:
            adjacency_matrix[node.process.process_data.ID] = [
                successor.process.process_data.ID for successor in node.successors
            ]
        return str(adjacency_matrix)

    def set_initial_marking(self):
        """
        Sets the initial marking of the process model. The initial marking is a node that has no predecessing nodes.

        Raises:
            ValueError: If no initial marking is found.
        """
        possible_starts = []
        for node in self.nodes:
            if not node.predecessors:
                possible_starts.append(node)
        if not possible_starts:
            raise ValueError("No initial marking found")
        self.current_marking = random.choice(possible_starts)

    def get_next_possible_processes(self) -> Optional[List[PROCESS_UNION]]:
        if not self.current_marking:
            self.set_initial_marking()
        possible_processes = [
            node.process
            for node in self.nodes
            if not node.marking
            and (not node.predecessors or all(n.marking for n in node.predecessors))
        ]
        return possible_processes

    def update_marking_from_transition(
        self, chosen_process: PROCESS_UNION
    ) -> None:
        chosen_node = [
            node for node in self.nodes if node.process == chosen_process
        ].pop()
        chosen_node.update_marking()
        self.current_marking = chosen_node

    def get_node_process_ids(self) -> List[str]:
        """
        Returns the process IDs of all nodes in the process model.

        Returns:
            List[str]: List of process IDs of all nodes in the process model.
        """
        return [node.process.process_data.ID for node in self.nodes]

    def add_node(
        self,
        process: PROCESS_UNION,
        successors: List[PROCESS_UNION],
        predecessors: List[PROCESS_UNION],
    ) -> None:
        """
        Adds a node to the process model. If the processes of the successors and predecessors are not in the process model, they are added as well, with for now empty lists of successors and predecessors.

        Args:
            process (PROCESS_UNION): Process that is represented by the node.
            successors (List[PROCESS_UNION]): List of successor processes.
            predecessors (List[PROCESS_UNION]): List of predecessor processes.
        """
        if not process.process_data.ID in self.get_node_process_ids():
            node = PrecendeGraphNode(process=process, successors=[], predecessors=[])
            self.nodes.append(node)
        else:
            node = [node for node in self.nodes if node.process == process].pop()

        successor_nodes: List[PrecendeGraphNode] = []
        for successor in successors:
            if successor.process_data.ID not in self.get_node_process_ids():
                self.add_node(successor, [], [])

            successor_nodes.append(
                [node for node in self.nodes if node.process == successor].pop()
            )
        node.successors = successor_nodes

        predecessor_nodes: List[PrecendeGraphNode] = []
        for predecessor in predecessors:
            if predecessor.process_data.ID not in self.get_node_process_ids():
                self.add_node(predecessor, [], [])
            predecessor_nodes.append(
                [node for node in self.nodes if node.process == predecessor].pop()
            )

        node.predecessors = predecessor_nodes

from prodsys.simulation.process import PROCESS_UNION


"""
In `prodsys` exist two different types of resources: production resources and transport resources. Production resources are resources that can perform processes on products. Transport resources are resources that can transport products from one location to another. Both types of resources are represented by the `Resource` class. The `Resource` class is an abstract base class and cannot be instantiated. Instead, the `ProductionResource` and `TransportResource` classes can be used to represent production resources and transport resources, respectively.

The following resources are available:

- `ProductionResource`: Class that represents a production resource.
- `TransportResource`: Class that represents a transport resource.
"""
from __future__ import annotations

from typing import List, Optional, Union
from uuid import uuid1

from abc import ABC

from pydantic import Field, conlist
from pydantic.dataclasses import dataclass

from prodsys.express import core

from prodsys.models import resource_data, queue_data
import prodsys

@dataclass
class Resource(ABC):
    """
    Abstract base class to represents a resource.

    Args:
        processes (List[process.PROCESS_UNION]): Processes of the resource.
        location (conlist(float, min_length=2, max_length=2)): Location of the resource.
        capacity (int): Capacity of the resource. Defaults to 1.
        states (Optional[List[state.STATE_UNION]], optional): States of the resource. Defaults to None.
        controller (resource_data.ControllerEnum, optional): Controller of the resource. Defaults to resource_data.ControllerEnum.PipelineController.
        control_policy (resource_data.ResourceControlPolicy, optional): Control policy of the resource. Defaults to resource_data.ResourceControlPolicy.FIFO.
        ID (str): ID of the resource.
    """
    processes: List[process.PROCESS_UNION]
    location: conlist(float, min_length=2, max_length=2) # type: ignore
    capacity: int = 1
    states: Optional[List[state.STATE_UNION]] = Field(default_factory=list)
    controller: resource_data.ControllerEnum = resource_data.ControllerEnum.PipelineController
    control_policy: Union[resource_data.ResourceControlPolicy, resource_data.TransportControlPolicy] = resource_data.ResourceControlPolicy.FIFO
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))


@dataclass
class ProductionResource(Resource, core.ExpressObject):
    """
    Class that represents a production resource.

    Args:
        processes (List[process.ProductionProcess]): Processes of the resource.
        location (conlist(float, min_length=2, max_length=2)): Location of the resource.
        capacity (int): Capacity of the resource. Defaults to 1.
        states (Optional[List[state.STATE_UNION]], optional): States of the resource. Defaults to None.
        controller (resource_data.ControllerEnum, optional): Controller of the resource. Defaults to resource_data.ControllerEnum.PipelineController.
        control_policy (resource_data.ResourceControlPolicy, optional): Control policy of the resource. Defaults to resource_data.ResourceControlPolicy.FIFO.
        queue_size (Optional[int], optional): Queue size of the resource. Defaults to 0 (infinte queue).
        ID (str): ID of the resource.

    Examples:
        Production resource with a capacity of 2 and 2 production processes:
        ``` py
        import prodsys.express as psx
        welding_time_model = psx.time_model_data.FunctionTimeModel(
            distribution_function="normal",
            location=20.0,
            scale=5.0,
        )
        screwing_time_model = psx.time_model_data.FunctionTimeModel(
            distribution_function="normal",
            location=10.0,
            scale=2.0,
        )
        welding_process = psx.process.ProductionProcess(
            time_model=welding_time_model
        )
        screwing_process = psx.process.ProductionProcess(
            time_model=screwing_time_model
        )
        psx.ProductionResource(
            processes=[welding_process, screwing_process],
            location=[10.0, 10.0]
            capacity=2
        )
        ```
    """
    processes: List[Union[process.ProductionProcess, process.CapabilityProcess]]
    control_policy: resource_data.ResourceControlPolicy = resource_data.ResourceControlPolicy.FIFO
    queue_size: Optional[int] = 0
    _input_queues: List[queue_data.QueueData] = Field(default_factory=list, init=False)
    _output_queues: List[queue_data.QueueData] = Field(default_factory=list, init=False)

    def to_model(self) -> resource_data.ProductionResourceData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            resource_data.ProductionResourceData: Data object of the express object.
        """
        resource =  resource_data.ProductionResourceData(
            ID=self.ID,
            description="",
            process_ids=[process.ID for process in self.processes],
            location=self.location,
            capacity=self.capacity,
            state_ids=[state.ID for state in self.states],
            controller=self.controller,
            control_policy=self.control_policy
        )
        self._input_queues, self._output_queues = prodsys.adapters.get_default_queues_for_resource(resource, self.queue_size)
        resource.input_queues = [q.ID for q in self._input_queues]
        resource.output_queues = [q.ID for q in self._output_queues]
        return resource   


@dataclass
class TransportResource(Resource, core.ExpressObject):
    """
    Class that represents a transport resource.

    Args:
        processes (List[process.TransportProcess]): Processes of the resource.
        location (conlist(float, min_length=2, max_length=2)): Location of the resource.
        capacity (int): Capacity of the resource. Defaults to 1.
        states (Optional[List[state.STATE_UNION]], optional): States of the resource. Defaults to None.
        controller (resource_data.ControllerEnum, optional): Controller of the resource. Defaults to resource_data.ControllerEnum.TransportController.
        control_policy (resource_data.TransportControlPolicy, optional): Control policy of the resource. Defaults to resource_data.TransportControlPolicy.FIFO.
        ID (str): ID of the resource.

    Examples:
        Transport resource with a capacity of 1:
        ``` py
        import prodsys.express as psx
        time_model = psx.time_model_data.ManhattanDistanceTimeModel(
            speed=1.0,
            reaction_time=1.0,
        )
        transport_process = psx.process.TransportProcess(
            time_model=time_model
        )
        psx.TransportResource(
            processes=[transport_process],
            location=[10.0, 10.0]
        )
        ```
    """
    processes: List[process.TransportProcess]
    location: conlist(float, min_length=2, max_length=2) = Field(default_factory=list) # type: ignore
    # capacity: int = 1
    # states: Optional[List[state.STATE_UNION]] = Field(default_factory=list)
    controller: resource_data.ControllerEnum = resource_data.ControllerEnum.TransportController
    control_policy: resource_data.TransportControlPolicy = resource_data.TransportControlPolicy.FIFO
    # ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    def __post_init__(self):
        if not self.location:
            self.location = [0.0, 0.0]

    def to_model(self) -> resource_data.TransportResourceData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            resource_data.TransportResourceData: Data object of the express object.
        """
        return resource_data.TransportResourceData(
            ID=self.ID,
            description="",
            process_ids=[process.ID for process in self.processes],
            location=self.location,
            capacity=self.capacity,
            state_ids=[state.ID for state in self.states],
            controller=self.controller,
            control_policy=self.control_policy,
        )

from prodsys.express import state, process




from typing import TYPE_CHECKING, Generator, List
from pydantic import BaseModel, Field
from prodsys.adapters import adapter
from prodsys.factories import process_factory, queue_factory, resource_factory, sink_factory
from prodsys.models import queue_data, auxiliary_data

from prodsys.simulation import sim, router
from prodsys.simulation import auxiliary, logger


from prodsys.models import source_data
class AuxiliaryFactory(BaseModel):
        """
        Factory class for creating and managing auxiliary objects in the production system.
        """

        env: sim.Environment
        process_factory: process_factory.ProcessFactory
        queue_factory: queue_factory.QueueFactory
        resource_factory: resource_factory.ResourceFactory
        sink_factory: sink_factory.SinkFactory
        auxiliaries: List[auxiliary.Auxiliary] = []
        event_logger: logger.EventLogger = Field(default=False, init=False)
        auxiliary_counter: int = 0

        class Config:
                arbitrary_types_allowed = True

        def create_auxiliary(self, adapter: adapter.ProductionSystemAdapter):
                """
                Create auxiliary objects based on the provided adapter's auxiliary data.

                Args:
                        adapter (adapter.ProductionSystemAdapter): The adapter containing auxiliary data.

                Returns:
                        None
                """
                for auxiliary_data in adapter.auxiliary_data:
                        for i, capacity_auxiliary_in_storage in enumerate(auxiliary_data.quantity_in_storages):
                                storage_id = auxiliary_data.storages[i]
                                storage = next((storage for storage in adapter.queue_data if storage.ID == storage_id), None)
                                for _ in range(capacity_auxiliary_in_storage):
                                        auxiliary = self.add_auxiliary(auxiliary_data, storage)
                                        auxiliary.auxiliary_info.log_create_auxiliary(
                                                resource=auxiliary.current_locatable, _product=auxiliary, event_time=self.env.now
                                        )

        def add_auxiliary(self, auxiliary_data: auxiliary_data.AuxiliaryData, storage: queue_data.QueueData):
                """
                Add a new auxiliary object to the factory.

                Args:
                        auxiliary_data (auxiliary_data.AuxiliaryData): The auxiliary data for the new object.
                        storage (queue_data.QueueData): The storage data for the new object's current location.

                Returns:
                        auxiliary.Auxiliary: The newly created auxiliary object.
                """
                values = {}
                auxiliary_data = auxiliary_data.model_copy(deep=True)
                auxiliary_data.ID = (str(auxiliary_data.auxiliary_type) + "_" + str(self.auxiliary_counter))
                values.update({"env": self.env, "data": auxiliary_data})
                transport_process = self.process_factory.get_process(auxiliary_data.transport_process)
                values.update({"transport_process": transport_process})
                relevant_processes = []
                for process_id in auxiliary_data.relevant_processes:
                        relevant_processes.append(self.process_factory.get_process(process_id))
                values.update({"relevant_processes": relevant_processes})
                relevant_transport_processes = []
                for transport_process_id in auxiliary_data.relevant_transport_processes:
                        relevant_transport_processes.append(self.process_factory.get_process(transport_process_id))
                values.update({"relevant_transport_processes": relevant_transport_processes})
                storage = self.queue_factory.get_queue(storage.ID)
                values.update({"storage": storage})
                router = self.get_router(source_data.RoutingHeuristic.FIFO)  # Add the routing_heuristic in auxiliary_data, like in source_data
                auxiliary_object = auxiliary.Auxiliary.model_validate(values)
                auxiliary_object.auxiliary_router = router
                auxiliary_object.current_locatable = storage
                auxiliary_object.init_got_free()

                if self.event_logger:
                        self.event_logger.observe_terminal_auxiliary_states(auxiliary_object)

                self.auxiliary_counter += 1
                self.auxiliaries.append(auxiliary_object)
                return auxiliary_object
        
        def place_auxiliaries_in_queues(self):
                """
                Place the auxiliary objects in the system.

                Returns:
                        None
                """
                for auxiliary in self.auxiliaries:
                        auxiliary.current_locatable.put(auxiliary.product_data)

        def get_auxiliary(self, ID: str) -> auxiliary.Auxiliary:
                """
                Get the auxiliary object with the specified ID.

                Args:
                        ID (str): The ID of the auxiliary object.

                Returns:
                        auxiliary.Auxiliary: The auxiliary object with the specified ID.

                Raises:
                        IndexError: If no auxiliary object with the specified ID is found.
                """
                return [s for s in self.auxiliaries if s.data.ID == ID].pop()

        def get_router(self, routing_heuristic: str):
                """
                Get the router object with the specified routing heuristic.

                Args:
                        routing_heuristic (str): The routing heuristic to use.

                Returns:
                        router.Router: The router object with the specified routing heuristic.
                """
                return router.Router(
                        self.resource_factory,
                        self.sink_factory,
                        self,
                        router.ROUTING_HEURISTIC[routing_heuristic],
                )
        

# from prodsys.factories import resource_factory, sink_factory
# AuxiliaryFactory.update_forward_refs()


"""
This module contains the adapters for the production system which are data containers for the data models defined in `prodsys.models`. The adapters are used in all algorithms that analyze or optimize a production system. 
The data structure of the adapter can also be used to integrate new algorithms into prodsys by utilizing its defined data structure. 
The adapters is the `prodsys.models` equivalent to the `prodsys.express.ProductionSystem` class and can be created from the express object. In contrast to the adapter, the express class nests the objects in a tree structure, which makes it easier to work with when instantiating a production system, but more complicated when reviewing the data itself.
The adapter comes with a data validation that ensures that entered data is syntatically, semantically and logically valid. 
"""
from prodsys.adapters.adapter import (
    get_default_queues_for_resource,
    get_default_queue_for_sink,
    get_default_queue_for_source,
    get_machines,
    get_transport_resources,
    get_set_of_IDs,
    get_possible_production_processes_IDs,
    add_default_queues_to_adapter,
    check_for_clean_compound_processes,
    ProductionSystemAdapter,
)
from prodsys.adapters.json_adapter import JsonProductionSystemAdapter



"""
This module contains the `prodsys.express` API with classes and functions to easily specify all parameters of a production system. This API is completely compatible with `prodsys.models` and every object can be converted to a data object from `prodsys.models` using the `to_model` method.

The express API is more convenient to use than the `prodsys.models` API because the express API nests the objects in a tree structure, which makes it easier to work with when instantiating production system programatically. 

However, the express API does not support all features of the `prodsys.models` API and saved data is more complicated to review because of the nesting.

Recommended is to use the API for programmatically specifying a production system but saving the data in the `prodsys.models` format with the `ProductionSystemAdapter`.

The following modules are available:

- `prodsys.express.core`: Contains the abstract base class for express objects.	
- `prodsys.express.product`: Contains classes to specify products.
- `prodsys.express.process`: Contains classes to specify processes.
- `prodsys.express.production_system`: Contains classes to specify a production system.
- `prodsys.express.resources`: Contains classes to specify resources.
- `prodsys.express.sink`: Contains classes to specify sinks.
- `prodsys.express.source`: Contains classes to specify sources.
- `prodsys.express.state`: Contains classes to specify states.
- `prodsys.express.time_model`: Contains classes to specify time models.

"""
from prodsys.express.queue import Queue
from prodsys.express.core import ExpressObject
from prodsys.express.time_model import (
    SampleTimeModel,
    ScheduledTimeModel,
    FunctionTimeModel,
    DistanceTimeModel,
    SequentialTimeModel,
    ManhattanDistanceTimeModel,
)
from prodsys.express.state import SetupState, BreakDownState, ProcessBreakdownState, ChargingState
from prodsys.express.node import Node
from prodsys.express.process import (
    ProductionProcess,
    CapabilityProcess,
    RequiredCapabilityProcess,
    TransportProcess,
    LinkTransportProcess,
)
from prodsys.express.resources import ProductionResource, TransportResource
from prodsys.express.product import Product
from prodsys.express.source import Source
from prodsys.express.sink import Sink
from prodsys.express.auxiliary import Auxiliary
from prodsys.express.production_system import ProductionSystem



"""
This package contains the simulation module. It is based on the SimPy package and uses the `prodsys.models` API for running the simulation. Objects are created with the `prodsys.factories`.

The simulation module contains the following modules:

- `prodsys.simulation.control`: Contains the logic for controlling the processes of resources in the simulation.
- `prodsys.simulation.logger`: Used for logging events in the simulation for later analysis.
- `prodsys.simulation.observer`: Can be used to observe the simulation and its processes while simulation.
- `prodsys.simulation.process_models`: Contains process models for products for the simulation.
- `prodsys.simulation.process`: All processes used in the simulation for product processing or transport.
- `prodsys.simulation.product`: Contains the logic for the product object in the simulation.
- `prodsys.simulation.request`: Has classes for representing requests of products to resources for processing or transport.
- `prodsys.simulation.resources`: Contains the logic for all resources in the simulation.
- `prodsys.simulation.router`: Contains the logic for routing products in the simulation.
- `prodsys.simulation.sim`: Simulation class for running the simulation.
- `prodsys.simulation.sink`: Contains the logic for sinks in the simulation.
- `prodsys.simulation.source`: Contains the logic for sources in the simulation.
- `prodsys.simulation.state`: Contains the logic for the state of resources in the simulation.
- `prodsys.simulation.store`: Contains the logic for product queues of resources in the simulation.
- `prodsys.simulation.time_model`: Contains the logic for time models in the simulation.


"""


from __future__ import annotations

from typing import List, Optional
from uuid import uuid1


from pydantic import Field, conlist
from pydantic.dataclasses import dataclass

from prodsys.express import core

from prodsys.models import sink_data, queue_data
import prodsys


@dataclass
class Sink(core.ExpressObject):
    """
    Class that represents a sink.

    Args:
        product (product.Product): Product of the sink.
        location (conlist(float, min_length=2, max_length=2)): Location of the sink.
        ID (str): ID of the sink.
    
    Attributes:
        _input_queues (List[queue_data.QueueData]): Input queues of the sink.
    
    Examples:
        Creation of a sink with a product and a location:
        ```py
        import prodsys.express as psx
        welding_time_model = psx.time_model_data.FunctionTimeModel(
            distribution_function="normal",
            location=20.0,
            scale=5.0,
        )
        welding_process_1 = psx.process.ProductionProcess(
            time_model=welding_time_model,
        )
        welding_process_2 = psx.process.ProductionProcess(
            time_model=welding_time_model,
        )
        transport_time_model = psx.time_model_data.ManhattenDistanceTimeModel(
            speed=10,
            reaction_time= 0.3
        )
        transport_process = psx.process.TransportProcess(
            time_model=transport_time_model,
        )
        product = psx.Product(
            processes=[welding_process_1, welding_process_2],
            transport_process=transport_process 
        )
        psx.Sink(
            product=product,
            location=[0.0, 0.0],
        )
        ```
    """
    product: product.Product
    location: conlist(float, min_length=2, max_length=2) # type: ignore
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    _input_queues: List[queue_data.QueueData] = Field(default_factory=list, init=False)

    def to_model(self) -> sink_data.SinkData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            sink_data.SinkData: Data object of the express object.
        """
        sink = sink_data.SinkData(
            ID=self.ID,
            description="",
            location=self.location,
            product_type=self.product.ID,
        )
        self._input_queues = [prodsys.adapters.get_default_queue_for_sink(sink)]
        sink.input_queues = [q.ID for q in self._input_queues]
        return sink
    
from prodsys.express import product



from __future__ import annotations
from hashlib import md5
from typing import List, Optional, TYPE_CHECKING
from pydantic import ConfigDict, conlist
from enum import Enum

from prodsys.models.core_asset import CoreAsset

if TYPE_CHECKING:
    from prodsys.adapters.adapter import ProductionSystemAdapter


class RoutingHeuristic(str, Enum):
    """
    Enum that represents the routing heuristic of a source.
    """
    random = "random"
    shortest_queue = "shortest_queue"
    FIFO = "FIFO"


class SourceData(CoreAsset):
    """
    Class that represents a source.

    Args:
        ID (str): ID of the source.
        description (str): Description of the source.
        location (List[float]): Location of the source. It has to be a list of length 2.
        product_type (str): Product type of the source.
        time_model_id (str): Time model ID of the source.
        router (RouterType): Router of the source.
        routing_heuristic (RoutingHeuristic): Routing heuristic of the source.
        output_queues (Optional[List[str]], optional): List of output queues of the source. Defaults to None.

    Examples:
        A source with ID "S1":
        ``` py
        import prodsys
        prodsys.source_data.SourceData(
            ID="S1",
            description="Source 1",
            location=[0.0, 0.0],
            product_type="Product_1",
            time_model_id="function_time_model_4",
            router="SimpleRouter",
            routing_heuristic="shortest_queue",
            output_queues=["SourceQueue"],
        )
    """
    location: conlist(float, min_length=2, max_length=2) # type: ignore
    product_type: str
    time_model_id: str
    routing_heuristic: RoutingHeuristic
    output_queues: List[str] = []
    
    def hash(self, adapter: ProductionSystemAdapter) -> str:
        """
        Returns a unique hash for the source considering its location, product type, time model, routing heuristic and output queues.

        Args:
            adapter (ProductionSystemAdapter): Adapter of the production system.

        Raises:
            ValueError: If the product, time model or output queue is not found in the adapter.

        Returns:
            str: Hash of the source.
        """
        for product in adapter.product_data:
            if product.product_type == self.product_type:
                product_hash = product.hash(adapter)
                break
        else:
            raise ValueError(f"Product with ID {self.product_type} not found for source {self.ID}.")

        for time_model in adapter.time_model_data:
            if time_model.ID == self.time_model_id:
                time_model_hash = time_model.hash()
                break
        else:
            raise ValueError(f"Time model with ID {self.time_model_id} not found for source {self.ID}.")
        
        output_queue_hashes = []
        for output_queue in self.output_queues:
            for queue in adapter.queue_data:
                if queue.ID == output_queue:
                    output_queue_hashes.append(queue.hash())
                    break
            else:
                raise ValueError(f"Queue with ID {output_queue} not found for source {self.ID}.")

        return md5(("".join([*map(str, self.location), product_hash, time_model_hash, self.routing_heuristic, *sorted(output_queue_hashes)])).encode("utf-8")).hexdigest()
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "S1",
                "description": "Source 1",
                "location": [0.0, 0.0],
                "product_type": "Product_1",
                "time_model_id": "function_time_model_4",
                "router": "SimpleRouter",
                "routing_heuristic": "shortest_queue",
                "output_queues": ["SourceQueue"],
            }
        ]
    })




import pandas as pd
from typing import List
import json
import plotly
import plotly.graph_objects as go
import numpy as np
from copy import copy


def read_optimization_results_file_to_df(filepath: str, label: str) -> pd.DataFrame:
    """
    Function reads the results of an optimization run from a json file and returns a pandas dataframe.

    Args:
        filepath (str): Filepath to the json file.
        label (str): Label to specify the utilized optimizer.

    Returns:
        pd.DataFrame: Dataframe containing the optimization results.
    """
    with open(filepath) as json_file:
        data = json.load(json_file)
    new_data = {}
    row_number = 1
    for generation, values in data.items():
        population_number = 0
        for individual, individual_values in values.items():
            ID = individual
            population_number += 1
            if "agg_fitness" in individual_values.keys():
                row_value = {
                    "Generation": int(generation),
                    "population_number": population_number,
                    "ID": ID,
                    "agg_fitness": individual_values["agg_fitness"],
                    "time": individual_values["time_stamp"],
                }
                for counter, KPI in enumerate(individual_values["fitness"]):
                    row_value.update({f"KPI_{counter}": KPI})
                new_data.update(
                    {
                        str(row_number): row_value,
                    }
                )
                row_number += 1
    df = pd.DataFrame(new_data)
    df = df.T
    df["optimizer"] = label
    if label == "anneal":
        df["agg_fitness"] = -1 * df["agg_fitness"]
    return df.copy()


def is_pareto_efficient_simple(costs: np.ndarray) -> np.ndarray:
    """
    Find the pareto-efficient points

    Args:
        costs (np.ndarray): An (n_points, n_costs) array

    Returns:
        np.ndarray: A (n_points, ) boolean array, indicating whether each point is Pareto efficient
    """
    is_efficient = np.ones(costs.shape[0], dtype=bool)
    for i, c in enumerate(costs):
        if is_efficient[i]:
            is_efficient[is_efficient] = np.any(
                costs[is_efficient] < c, axis=1
            )  # Keep any point with a lower cost
            is_efficient[i] = True  # And keep self
    return is_efficient


# Faster than is_pareto_efficient_simple, but less readable.
def is_pareto_efficient(costs):
    """
    Find the pareto-efficient points

    Args:
        costs (np.ndarray): An (n_points, n_costs) array

    Returns:
        np.ndarray: A (n_points, ) boolean array, indicating whether each point is Pareto efficient
    """
    is_efficient = np.arange(costs.shape[0])
    n_points = costs.shape[0]
    next_point_index = 0  # Next index in the is_efficient array to search for
    while next_point_index < len(costs):
        nondominated_point_mask = np.any(costs < costs[next_point_index], axis=1)
        nondominated_point_mask[next_point_index] = True
        is_efficient = is_efficient[nondominated_point_mask]  # Remove dominated points
        costs = costs[nondominated_point_mask]
        next_point_index = np.sum(nondominated_point_mask[:next_point_index]) + 1
    is_efficient_mask = np.zeros(n_points, dtype=bool)
    is_efficient_mask[is_efficient] = True
    return is_efficient_mask


def get_pareto_solutions_from_result_files(
    file_path: str,
) -> List[str]:
    """
    Analyses an optimization result file and returns the IDs of the pareto efficient solutions.

    Args:
        file_path (str): Filepath of the json file containing the optimization results.

    Returns:
        List[str]: List of IDs of the pareto efficient solutions.
    """
    df = read_optimization_results_file_to_df(file_path, label="optimizer")
    df = df.drop_duplicates(
        subset=["agg_fitness", "cost", "throughput", "wip", "optimizer"]
    )
    df["cost"] = df["cost"].astype(float)
    df["agg_fitness"] = df["agg_fitness"].astype(float)
    # df = df.loc[df["wip"] < 150]
    # df = df.loc[df["throughput"] > 100]
    # # df = df.loc[df["throughput_time"] < 100]
    columns = ["cost", "throughput", "wip"]
    # columns = ['throughput_time', 'throughput']
    df_for_pareto = df[columns].copy()
    df_for_pareto["throughput"] = -df_for_pareto["throughput"]
    is_efficient = is_pareto_efficient(df_for_pareto.values)
    df["is_efficient"] = is_efficient
    return df.loc[df["is_efficient"]]["ID"].to_list()



from enum import Enum
from pydantic import BaseModel, ConfigDict
from typing import Literal, Union, Optional, List, Tuple, TYPE_CHECKING

from prodsys.models.performance_indicators import (
    KPIEnum,
    KPILevelEnum,
    KPI_UNION,
)


class Event(BaseModel):
    """
    Class that represents an event in the event log.

    Args:
        time (float): Time of the event.
        resource (str): Resource of the event.
        state (str): State of the event.
        state_type (str): State type of the event, representing a state.StateTypeEnum.
        activity (str): Activity of the event, representing a state.StateEnum.
        product (Optional[str], optional): Product of the event. Defaults to None.
        expected_end_time (Optional[float], optional): Expected end time of the event. Defaults to None.
        target_location (Optional[str], optional): Target location of the event. Defaults to None.
    """

    time: float
    resource: str
    state: str
    state_type: str
    activity: str
    product: Optional[str] = None
    expected_end_time: Optional[float] = None
    target_location: Optional[str] = None

    model_config = ConfigDict(json_schema_extra={
        "examples": [
            {
                "time": 12.0,
                "resource": "R1",
                "state": "P1",
                "state_type": "Production",
                "activity": "start state",
                "product": "Product_1_12",
                "expected_end_time": 24.3,
                "target_location": None,
            }, 
            {
                "time": 24.3,
                "resource": "R1",
                "state": "P1",
                "state_type": "Production",
                "activity": "end state",
                "product": "Product_1_12",
                "expected_end_time": None,
                "target_location": "L1",
            },
        ]
    })


class Performance(BaseModel):
    """
    Class that represents the performance of a simulation run.

    Args:
        event_log (List[Event]): Event log of the simulation run.
        kpis (List[KPI_UNION]): List of KPIs of the simulation run.
    """

    model_config = ConfigDict(json_schema_extra={
        "examples": [
            {
                "event_log": Event.model_config["json_schema_extra"]["examples"],
                "kpis": [
                    {
                        "name": "throughput",
                        "target": "max",
                        "weight": 1,
                        "value": 4.32,
                        "context": ["system", "product_type"],
                        "product_type": "ProductType_1",
                    },
                    {
                        "name": "WIP",
                        "target": "min",
                        "weight": 1,
                        "value": 121,
                        "context": ["system", "product_type"],
                        "product_type": "ProductType_1",
                    },
                ],
            }
        ]
    })

    event_log: List[Event]
    kpis: List[KPI_UNION]

    def get_kpi_for_context(self, context: Tuple[KPILevelEnum, ...]) -> List[KPI_UNION]:
        return [kpi for kpi in self.kpis if context == kpi.context]

    def get_kpi_for_name(self, name: KPIEnum) -> List[KPI_UNION]:
        return [kpi for kpi in self.kpis if name == kpi.name]

    def get_kpi_for_context_and_name(
        self, context: Tuple[KPILevelEnum, ...], name: KPIEnum
    ) -> List[KPI_UNION]:
        return [kpi for kpi in self.kpis if context == kpi.context and name == kpi.name]



from __future__ import annotations

from typing import List, TYPE_CHECKING

from pydantic import BaseModel, TypeAdapter
from prodsys.models.time_model_data import TIME_MODEL_DATA
from prodsys.simulation.time_model import TIME_MODEL, TimeModel

if TYPE_CHECKING:
    from prodsys.adapters import adapter

class TimeModelFactory(BaseModel):
    """
    Factory class that creates and stores `prodsys.simulation` time model objects based on the given time model data according to `prodsys.models.time_model_data.TIME_MODEL_DATA`.

    Returns:
        _type_: _description_
    """
    time_model_data: List[TIME_MODEL_DATA] = []
    time_models: List[TIME_MODEL] = []

    def create_time_models(self, adapter: adapter.ProductionSystemAdapter):
        """
        Creates time model objects based on the given adapter.

        Args:
            adapter (adapter.ProductionSystemAdapter): Adapter that contains the time model data.
        """
        for time_model_data in adapter.time_model_data:
            self.time_models.append(TypeAdapter(TIME_MODEL).validate_python({"time_model_data": time_model_data})
            )

    def get_time_models(self, IDs: List[str]) -> List[TimeModel]:
        """
        Returns a list of time model objects with the given IDs.

        Args:
            IDs (List[str]): List of IDs that is used to filter the time model objects.

        Returns:
            List[time_model.TimeModel]: List of time model objects with the given IDs.
        """
        return [tm for tm in self.time_models if tm.time_model_data.ID in IDs]

    def get_time_model(self, ID: str) -> TimeModel:
        """
        Returns a time model object with the given ID.

        Args:
            ID (str): ID that is used to filter the time model objects.

        Returns:
            time_model.TimeModel: Time model object with the given ID.
        """
        return [tm for tm in self.time_models if tm.time_model_data.ID == ID].pop()



"""
The `state_data` module contains the `prodsys.models` classes to represent the states that resources 
can be in during a simulation.

The following states are possible:

- `BreakDownStateData`: A state that makes a resource unavailable for a certain time.
- `ProcessBreakDownStateData`: A state that makes a process unavailable for a certain time but other processes can still be performed.
- `SetupStateData`: A state that represents the time needed to change the process of a resource.
- `ProductionStateData`: A state that represents the time needed to process a product.
- `TransportStateData`: A state that represents the time needed to transport a product.	    
"""
from __future__ import annotations

from hashlib import md5
from enum import Enum
from typing import Literal, Union, TYPE_CHECKING

from pydantic import ConfigDict

from prodsys.models.core_asset import CoreAsset

if TYPE_CHECKING:
    from prodsys.adapters.adapter import ProductionSystemAdapter


class StateTypeEnum(str, Enum):
    """
    Enum that represents the different kind of states.

    - BreakDownState: A state that makes a resource unavailable for a certain time.
    - ProductionState: A state that represents the time needed to process a product.
    - TransportState: A state that represents the time needed to transport a product.
    - SetupState: A state that represents the time needed to change the process of a resource.
    - ProcessBreakDownState: A state that makes a process unavailable for a certain time but other processes can still be performed.
    """

    BreakDownState = "BreakDownState"
    ProductionState = "ProductionState"
    TransportState = "TransportState"
    SetupState = "SetupState"
    ProcessBreakDownState = "ProcessBreakDownState"
    ChargingState = "ChargingState"


class StateData(CoreAsset):
    """
    Class that represents a state.

    Args:
        ID (str): ID of the state.
        description (str): Description of the state.
        time_model_id (str): Time model ID of the state.
        type (StateTypeEnum): Type of the state.
    """

    time_model_id: str
    type: Literal[
        StateTypeEnum.BreakDownState,
        StateTypeEnum.ProductionState,
        StateTypeEnum.TransportState,
        StateTypeEnum.SetupState,
    ]

    def hash(self, adapter: ProductionSystemAdapter) -> str:
        """
        Returns a unique hash of the state considering the time model and the type of the state. Can be used to compare states for equal functionality.

        Args:
            adapter (ProductionSystemAdapter): Adapter to access the data of the state.

        Raises:
            ValueError: if the time model is not found in the adapter.

        Returns:
            str: hash of the state.
        """
        time_model_hash = ""
        for time_model in adapter.time_model_data:
            if time_model.ID == self.time_model_id:
                time_model_hash = time_model.hash()
                break
        else:
            raise ValueError(f"Time model with ID {self.time_model_id} not found for state {self.ID}.")
        return md5(("".join([self.type, time_model_hash])).encode("utf-8")).hexdigest()
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "state_1",
                "description": "State data for state_1",
                "time_model_id": "time_model_1",
                "type": "ProductionState",
            }
        ]
    })


class BreakDownStateData(StateData):
    """
    Class that represents a breakdown state.

    Args:
        ID (str): ID of the state.
        description (str): Description of the state.
        time_model_id (str): Time model ID of the state. Specifies the time interval between breakdowns.
        type (StateTypeEnum): Type of the state.
        repair_time_model_id (str): Time model ID of the repair time.

    Examples:
        Breakdown state with a function time model:
        ``` py  
        import prodsys
        prodsys.state_data.BreakDownStateData(
            ID="Breakdownstate_1",
            description="Breakdown state machine 1",
            time_model_id="function_time_model_5",
            repair_time_model_id="function_time_model_8",
        )
        ```
    """

    type: Literal[StateTypeEnum.BreakDownState]
    repair_time_model_id: str

    def hash(self, adapter: ProductionSystemAdapter) -> str:
        """
        Returns a unique hash of the state considering the time model and the repair time model. Can be used to compare states for equal functionality.

        Args:
            adapter (ProductionSystemAdapter): Adapter to access the data of the state.

        Raises:
            ValueError: if the repair time model is not found.

        Returns:
            str: hash of the state.
        """
        base_class_hash = super().hash(adapter)
        repair_time_model_hash = ""

        for repair_time_model in adapter.time_model_data:
            if repair_time_model.ID == self.repair_time_model_id:
                repair_time_model_hash = repair_time_model.hash()
                break
        else:
            raise ValueError(f"Repair time model with ID {self.repair_time_model_id} not found for state {self.ID}.")

        return md5(("".join([base_class_hash, repair_time_model_hash])).encode("utf-8")).hexdigest()  
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "Breakdownstate_1",
                "description": "Breakdown state machine 1",
                "time_model_id": "function_time_model_5",
                "type": "BreakDownState",
                "repair_time_model_id": "function_time_model_8",
            }
        ]
    })


class ProcessBreakDownStateData(StateData):
    """
    Class that represents a process breakdown state. It is a breakdown state that is connected to a process. Other processes can still be executed while the process breakdown state is activen.

    Args:
        ID (str): ID of the state.
        description (str): Description of the state.
        time_model_id (str): Time model ID of the state.
        type (StateTypeEnum): Type of the state.
        repair_time_model_id (str): Time model ID of the repair time.
        process_id (str): ID of the process that is broken down.

    Examples:
        Process breakdown state with a function time model:
        ``` py
        import prodsys
        prodsys.state_data.ProcessBreakDownStateData(
            ID="ProcessBreakDownState_1",
            description="Process Breakdown state machine 1",
            time_model_id="function_time_model_7",
            repair_time_model_id="function_time_model_8",
            process_id="P1",
        )
        ```
    """

    type: Literal[StateTypeEnum.ProcessBreakDownState]
    repair_time_model_id: str
    process_id: str

    def hash(self, adapter: ProductionSystemAdapter) -> str:
        """
        Returns a unique hash of the state considering the time model, process and repair time model. Can be used to compare states for equal functionality.

        Args:
            adapter (ProductionSystemAdapter): Adapter to access the data of the state.

        Raises:
            ValueError: if the process or repair time model is not found.

        Returns:
            str: hash of the state.
        """
        base_class_hash = super().hash(adapter)
        process_hash = ""
        repair_time_model_hash = ""

        for process in adapter.process_data:
            if process.ID == self.process_id:
                process_hash = process.hash(adapter)
                break
        else:
            raise ValueError(f"Process with ID {self.process_id} not found for process breakdown state {self.ID}.")

        for repair_time_model in adapter.time_model_data:
            if repair_time_model.ID == self.repair_time_model_id:
                repair_time_model_hash = repair_time_model.hash()
                break
        else:
            raise ValueError(f"Repair time model with ID {self.repair_time_model_id} not found for process breakdown state {self.ID}.")
        

        return md5(("".join([base_class_hash, process_hash, repair_time_model_hash])).encode("utf-8")).hexdigest()
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "ProcessBreakDownState_1",
                "description": "Process Breakdown state machine 1",
                "time_model_id": "function_time_model_7",
                "type": "ProcessBreakDownState",
                "process_id": "P1",
                "repair_time_model_id": "function_time_model_8",
            }
        ]
    })


class ProductionStateData(StateData):
    """
    Class that represents a production state. By undergoing a production state, the product is processed and continues its process model. Production states don't have to be initialized because they are automatically created when a process is added to a resource.

    Args:
        ID (str): ID of the state.
        description (str): Description of the state.
        time_model_id (str): Time model ID of the state.
        type (StateTypeEnum): Type of the state.

    """

    type: Literal[StateTypeEnum.ProductionState]

    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "ProductionState_1",
                "description": "Production state machine 1",
                "time_model_id": "function_time_model_1",
                "type": "ProductionState",
            }
        ]
    })

class TransportStateData(StateData):
    """
    Class that represents a transport state. By undergoing a transport state, the product is transported and its position is changed. Transport states don't have to be initialized because they are automatically created when a transport process is added to a resource.

    Args:
        ID (str): ID of the state.
        description (str): Description of the state.
        time_model_id (str): Time model ID of the state.
        type (StateTypeEnum): Type of the state.
    """

    type: Literal[StateTypeEnum.TransportState]

    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "TransportState_1",
                "description": "Transport state machine 1",
                "time_model_id": "function_time_model_3",
                "type": "TransportState",
            }
        ]
    })


class SetupStateData(StateData):
    """
    Class that represents a setup state. By undergoing a setup state, the process is setup.

    Args:
        ID (str): ID of the state.
        description (str): Description of the state.
        time_model_id (str): Time model ID of the state.
        type (StateTypeEnum): Type of the state.
        origin_setup (str): ID of the origin process for the setup.
        target_setup (str): ID of the target process for the setup.

    Examples:
        Setup state with a function time model:
        ``` py
        import prodsys
        prodsys.state_data.SetupStateData(
            ID="Setup_State_2",
            description="Setup state machine 2",
            time_model_id="function_time_model_2",
            origin_setup="P2",
            target_setup="P1",
        )
        ```
    """

    type: Literal[StateTypeEnum.SetupState]
    origin_setup: str
    target_setup: str

    def hash(self, adapter: ProductionSystemAdapter) -> str:
        """
        Returns a unique hash of the state considering the time model, origin and target setup process. Can be used to compare states for equal functionality.

        Args:
            adapter (ProductionSystemAdapter): Adapter to access the data of the state.

        Raises:
            ValueError: if the origin or target setup process is not found.

        Returns:
            str: hash of the state.
        """
        base_class_hash = super().hash(adapter)
        setup_process_hashes = []

        for process_id in [self.origin_setup, self.target_setup]:
            for process in adapter.process_data:
                if process.ID == process_id:
                    setup_process_hashes.append(process.hash(adapter))
                break
            else:
                raise ValueError(f"Process with ID {process_id} not found for setup state {self.ID}.")

        return md5(("".join([base_class_hash] + setup_process_hashes)).encode("utf-8")).hexdigest()
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "Setup_State_2",
                "description": "Setup state machine 2",
                "time_model_id": "function_time_model_2",
                "type": "SetupState",
                "origin_setup": "P2",
                "target_setup": "P1",
            }
        ]
    })


class ChargingStateData(StateData):
    """
    Class that represents a battery powered resource.

    Args:
        ID (str): ID of the state.
        description (str): Description of the state.
        time_model_id (str): Time model ID of the state. Specifies the time for charging the AGV fully.
        type (StateTypeEnum): Type of the state.
        battery_time_mdoel_id (str): Time model ID of the battery time available for operation of the resource.

    Examples:
        Breakdown state with a function time model:
        ``` py  
        import prodsys
        prodsys.state_data.ChargingStateData(
            ID="ChargingState_1",
            description="Charging state machine 1",
            time_model_id="function_time_model_5",
            repair_time_model_id="function_time_model_8",
        )
        ```
    """

    type: Literal[StateTypeEnum.ChargingState]
    battery_time_model_id: str

    def hash(self, adapter: ProductionSystemAdapter) -> str:
        """
        Returns a unique hash of the state considering the time model and the repair time model. Can be used to compare states for equal functionality.

        Args:
            adapter (ProductionSystemAdapter): Adapter to access the data of the state.

        Raises:
            ValueError: if the repair time model is not found.

        Returns:
            str: hash of the state.
        """
        base_class_hash = super().hash(adapter)
        charging_time_model_hash = ""

        for charging_time_model in adapter.time_model_data:
            if charging_time_model.ID == self.battery_time_model_id:
                charging_time_model_hash = charging_time_model.hash()
                break
        else:
            raise ValueError(f"Battery time model with ID {self.battery_time_model_id} not found for state {self.ID}.")

        return md5(("".join([base_class_hash, charging_time_model_hash])).encode("utf-8")).hexdigest()  
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "ChargingState_1",
                "description": "Charging state machine 1",
                "time_model_id": "function_time_model_5",
                "type": "ChargingState",
                "battery_time_mdoel_id": "function_time_model_8",
            }
        ]
    })

STATE_DATA_UNION = Union[
    BreakDownStateData,
    ChargingStateData,
    ProductionStateData,
    TransportStateData,
    SetupStateData,
    ProcessBreakDownStateData,
]



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional
from pydantic import BaseModel
from pydantic import TypeAdapter
from warnings import warn


from prodsys.adapters import adapter

from prodsys.models import (
    auxiliary_data,
    product_data,
    queue_data,
    resource_data,
    node_data,
    time_model_data,
    state_data,
    processes_data,
    sink_data,
    source_data,
)

def load_json(file_path: str) -> dict:
    with open(file_path, "r", encoding="utf-8") as json_file:
        data = json.load(json_file)
    return data

class JsonProductionSystemAdapter(adapter.ProductionSystemAdapter):
    """
    JsonProductionSystemAdapter is a class that implements the abstract class ProductionSystemAdapter and allows to read and write data from and to a json file.

    Args:
        ID (str, optional): ID of the production system. Defaults to "".
        seed (int, optional): Seed for the random number generator used in simulation. Defaults to 0.
        time_model_data (List[time_model_data.TIME_MODEL_DATA], optional): List of time models used by the entities in the production system. Defaults to [].
        state_data (List[state_data.STATE_DATA_UNION], optional): List of states used by the resources in the production system. Defaults to [].
        process_data (List[processes_data.PROCESS_DATA_UNION], optional): List of processes required by products and provided by resources in the production system. Defaults to [].
        queue_data (List[queue_data.QueueData], optional): List of queues used by the resources, sources and sinks in the production system. Defaults to [].
        resource_data (List[resource_data.RESOURCE_DATA_UNION], optional): List of resources in the production system. Defaults to [].
        node_data (List[resource_data.NodeData], optional): List of nodes in the production system. Defaults to [].
        product_data (List[product_data.ProductData], optional): List of products in the production system. Defaults to [].
        sink_data (List[sink_data.SinkData], optional): List of sinks in the production system. Defaults to [].
        source_data (List[source_data.SourceData], optional): List of sources in the production system. Defaults to [].
        scenario_data (Optional[scenario_data.ScenarioData], optional): Scenario data of the production system used for optimization. Defaults to None.
        valid_configuration (bool, optional): Indicates if the configuration is valid. Defaults to True.
        reconfiguration_cost (float, optional): Cost of reconfiguration in a optimization scenario. Defaults to 0.
    """

    def read_data_old(self, file_path: str, scenario_file_path: Optional[str] = None):
        """
        Reads the data from the given file path and scenario file path.

        Args:
            file_path (str): File path for the production system configuration
            scenario_file_path (Optional[str], optional): File path for the scenario data. Defaults to None.
        """
        warn("This method is deprecated. Use read_data instead.", DeprecationWarning)
        data = load_json(file_path=file_path)
        self.seed = data["seed"]
        self.time_model_data = self.create_objects_from_configuration_data_old(
            data["time_models"], time_model_data.TIME_MODEL_DATA
        )
        self.state_data = self.create_objects_from_configuration_data_old(
            data["states"], state_data.STATE_DATA_UNION
        )
        self.process_data = self.create_objects_from_configuration_data_old(
            data["processes"], processes_data.PROCESS_DATA_UNION
        )

        self.queue_data = self.create_objects_from_configuration_data_old(data["queues"], queue_data.QueueData)
        self.resource_data = self.create_objects_from_configuration_data_old(data["resources"], resource_data.RESOURCE_DATA_UNION)
        self.product_data = self.create_objects_from_configuration_data_old(data["products"], product_data.ProductData)
        self.node_data = self.create_objects_from_configuration_data(data["links"], node_data.NodeData)
        self.sink_data = self.create_objects_from_configuration_data_old(data["sinks"], sink_data.SinkData)
        self.source_data = self.create_objects_from_configuration_data_old(data["sources"], source_data.SourceData)
        if scenario_file_path:
            self.read_scenario(scenario_file_path)

    def read_data(self, file_path: str, scenario_file_path: Optional[str] = None):
        """
        Reads the data from the given file path and scenario file path.

        Args:
            file_path (str): File path for the production system configuration
            scenario_file_path (Optional[str], optional): File path for the scenario data. Defaults to None.
        """
        data = load_json(file_path=file_path)
        if "ID" in data:
            self.ID = data["ID"]
        if "seed" in data:
            self.seed = data["seed"]
        else:
            self.seed = 0
        self.time_model_data = self.create_objects_from_configuration_data(
            data["time_model_data"], time_model_data.TIME_MODEL_DATA
        )
        self.state_data = self.create_objects_from_configuration_data(
            data["state_data"], state_data.STATE_DATA_UNION
        )
        self.process_data = self.create_objects_from_configuration_data(
            data["process_data"], processes_data.PROCESS_DATA_UNION
        )
        self.queue_data = self.create_objects_from_configuration_data(data["queue_data"], queue_data.QueueData)
        self.resource_data = self.create_objects_from_configuration_data(data["resource_data"], resource_data.RESOURCE_DATA_UNION)
        self.product_data = self.create_objects_from_configuration_data(data["product_data"], product_data.ProductData)
        self.sink_data = self.create_objects_from_configuration_data(data["sink_data"], sink_data.SinkData)
        if "node_data" in data:
            self.node_data = self.create_objects_from_configuration_data(data["node_data"], node_data.NodeData)
        if "auxiliary_data" in data:
            self.auxiliary_data = self.create_objects_from_configuration_data(data["auxiliary_data"], auxiliary_data.AuxiliaryData)
        self.source_data = self.create_objects_from_configuration_data(data["source_data"], source_data.SourceData)
        if scenario_file_path:
            self.read_scenario(scenario_file_path)
    
    def create_objects_from_configuration_data_old(
        self, configuration_data: Dict[str, Any], type
    ):  
        warn("This method is deprecated. Use create_objects_from_configuration_data instead.", DeprecationWarning)
        objects = []
        for values in configuration_data.values():
            objects.append(TypeAdapter(type).validate_python(values))
        return objects
    
    def create_objects_from_configuration_data(
        self, configuration_data: List[Any], type
    ):  
        objects = []
        for values in configuration_data:
            objects.append(TypeAdapter(type).validate_python(values))
        return objects

    def write_data(self, file_path: str):
        """
        Writes the data to the given file path.

        Args:
            file_path (str): File path for the production system configuration
        """
        with open(file_path, "w") as json_file:
            json_file.write(self.model_dump_json(indent=4))
    
    def write_scenario_data(self, file_path: str) -> None:
        """
        Writes the scenario data to the given file path.

        Args:
            file_path (str): File path for the scenario data.
        """
        with open(file_path, "w") as json_file:
            json_file.write(self.scenario_data.model_dump_json(indent=4))


from __future__ import annotations

from typing import List, TYPE_CHECKING

from pydantic import BaseModel, ConfigDict



if TYPE_CHECKING:
    from prodsys.simulation import resources, state
    from prodsys.factories import resource_factory


class ProcessObservation(BaseModel):
    process: str
    product: str
    activity: state.StateEnum
    state: state.StateTypeEnum

class QueueObservation(BaseModel):
    product: str
    activity: str
    process: str
    next_resource: str
    waiting_since: float

def observe_processes(resource: resources.Resource) -> List[ProcessObservation]:
    process_observations = []
    for production_state in resource.production_states:
        if production_state.process:
            observation = ProcessObservation(
                process=production_state.state_info.ID,
                product=production_state.state_info._product_ID,
                activity=production_state.state_info._activity,
                state=production_state.state_info._state_type
            )
            process_observations.append(observation)
    return process_observations

def observe_input_queue(resource: resources.Resource, product_factory: product_factory.ProductFactory) -> List[QueueObservation]:
    queue_observation = []
    for queue in resource.input_queues:
        for product_data in queue.items:
            product = product_factory.get_product(product_data.ID)

            production_process_info = QueueObservation(
                product=product_data.ID,
                activity="waiting",
                process=product.next_prodution_process.process_data.ID,
                next_resource=product.current_locatable.data.ID,
                waiting_since=product.product_info.event_time
            )
            queue_observation.append(production_process_info)

    return queue_observation

def observe_output_queue(resource: resources.Resource, product_factory: product_factory.ProductFactory) -> List[QueueObservation]:
    queue_observation = []
    for queue in resource.output_queues:
        for product_data in queue.items:
            product = product_factory.get_product(product_data.ID)

            production_process_info = QueueObservation(
                product=product_data.ID,
                activity="waiting",
                process=product.next_prodution_process,
                next_resource=product.current_locatable,
                waiting_since=product.product_info.event_time
            )
            queue_observation.append(production_process_info)

    return queue_observation

class ResourceAvailableObservation(BaseModel):
    available: bool

def observe_resource_available(resource: resources.Resource) -> ResourceAvailableObservation:
    return ResourceAvailableObservation(available=resource.active.triggered)
    

class ResourceObserver(BaseModel):
    resource_factory: resource_factory.ResourceFactory
    product_factory: product_factory.ProductFactory
    resource: resources.Resource

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def observe_processes(self) -> List[ProcessObservation]:
        return observe_processes(self.resource)
    
    def observe_input_queue(self) -> List[QueueObservation]:
        return observe_input_queue(self.resource, self.product_factory)
    
    def observe_output_queue(self) -> List[QueueObservation]:
        return observe_output_queue(self.resource, self.product_factory)
    
    def observe_resource_available(self) -> ResourceAvailableObservation:
        return observe_resource_available(self.resource)
    
from prodsys.factories import resource_factory, product_factory
from prodsys.simulation import resources, state



from __future__ import annotations

from typing import List, TYPE_CHECKING

from pydantic import BaseModel, ConfigDict, Field

from prodsys.simulation import sim, store
from prodsys.models import sink_data

if TYPE_CHECKING:
    from prodsys.simulation import product


class Sink(BaseModel):
    """
    Class that represents a sink.

    Args:
        env (sim.Environment): The simulation environment.
        data (sink_data.SinkData): The sink data.
        product_factory (product_factory.ProductFactory): The product factory.
        input_queues (List[store.Queue], optional): The input queues. Defaults to [].
    """
    env: sim.Environment
    data: sink_data.SinkData
    product_factory: product_factory.ProductFactory
    input_queues: List[store.Queue] = Field(default_factory=list, init=False)

    model_config=ConfigDict(arbitrary_types_allowed=True)

    def add_input_queues(self, input_queues: List[store.Queue]):
        """
        Adds input queues to the sink.

        Args:
            input_queues (List[store.Queue]): The input queues.
        """
        self.input_queues.extend(input_queues)

    def get_location(self) -> List[float]:
        """
        Returns the location of the sink.

        Returns:
            List[float]: The location. Has to be a list of length 2.
        """
        return self.data.location

    def get_input_queue_length(self) -> int:
        """
        Returns total number of items in all input_queues. Defaults to 0 for a sink.

        Returns:
            int(0)
        """
        return 0

    def get_output_queue_length(self) -> int:
        """
        Returns zero. Needed for transport control policies that use the target locations output queue for request ordering.
        Returns:
            int(0)
        """
        return 0
    
    def register_finished_product(self, product: product.Product):
        """
        Registers a finished product when it reaches the sink.

        Args:
            product (product.Product): The finished product.
        """
        self.product_factory.register_finished_product(product)

from prodsys.factories import product_factory    




import json
import time
import warnings
import logging

from prodsys.optimization.optimization import evaluate
from prodsys.optimization.adapter_manipulation import crossover, mutation, random_configuration, random_configuration_with_initial_solution
from prodsys.optimization.util import document_individual
logger = logging.getLogger(__name__)

warnings.filterwarnings("ignore", category=RuntimeWarning)

from deap import algorithms, base, creator, tools
from pydantic import BaseModel, ConfigDict, Field

from prodsys.simulation import sim
from prodsys import adapters
from prodsys.optimization.util import (
    get_weights,
    check_breakdown_states_available,
    create_default_breakdown_states,
)
from prodsys.util.util import set_seed, read_initial_solutions, run_from_ipython
from prodsys import optimization



if run_from_ipython():
    from multiprocessing.pool import ThreadPool as Pool
else:
    from multiprocessing.pool import Pool

sim.VERBOSE = 1


creator.create("FitnessMax", base.Fitness, weights=(1, 1, 1))  # als Tupel
creator.create("Individual", list, fitness=creator.FitnessMax)

class EvolutionaryAlgorithmHyperparameters(BaseModel):
    """
    Hyperparameters for configuration optimization using an evolutionary algorithm.

    Args:
        seed (int): Seed for the random number generator.
        number_of_generations (int): Number of generations to run the algorithm.
        population_size (int): Number of individuals in each generation.
        mutation_rate (float): Probability of mutating an individual.
        crossover_rate (float): Probability of crossover between two individuals.
        number_of_seeds (int): Number of seeds to use for simulation.
        number_of_processes (int): Number of processes to use for parallelization.
    """

    seed: int = Field(0, description="Seed for the random number generator.")
    number_of_generations: int = 10
    population_size: int = 10
    mutation_rate: float = 0.1
    crossover_rate: float = 0.1
    number_of_seeds: int = 1
    number_of_processes: int = 1

    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "seed": 0,
                "number_of_generations": 10,
                "population_size": 10,
                "mutation_rate": 0.1,
                "crossover_rate": 0.1,
                "number_of_seeds": 1,
                "number_of_processes": 1,
            },
        ]
    })


def register_functions_in_toolbox(
    base_configuration: adapters.JsonProductionSystemAdapter,
    solution_dict: dict,
    performances: dict,
    weights: tuple,
    initial_solutions_folder: str,
    hyper_parameters: EvolutionaryAlgorithmHyperparameters,
    full_save_solutions_folder: str="",
):
    creator.create("FitnessMax", base.Fitness, weights=weights)  # als Tupel
    creator.create("Individual", list, fitness=creator.FitnessMax)
    toolbox = base.Toolbox()
    if initial_solutions_folder:
        initial_solutions = read_initial_solutions(
            initial_solutions_folder, base_configuration
        )
        toolbox.register(
            "random_configuration",
            random_configuration_with_initial_solution,
            initial_solutions,
        )
    else:
        toolbox.register(
            "random_configuration", random_configuration, base_configuration
        )
    toolbox.register(
        "individual",
        tools.initRepeat,
        creator.Individual,
        toolbox.random_configuration,
        n=1,
    )

    # Startpopulation erzeugen
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register(
        "evaluate",
        evaluate,
        base_configuration,
        solution_dict,
        performances,
        hyper_parameters.number_of_seeds, 
        full_save_solutions_folder
    )
    toolbox.register("mate", crossover)
    toolbox.register("mutate", mutation)

    toolbox.register("select", tools.selNSGA2)

    return toolbox


def save_population_results(
    population, fitnesses, solution_dict, performances, save_folder, start
):
    generation_performances = []

    for ind, fit in zip(population, fitnesses):
        document_individual(solution_dict, save_folder, ind)
        ind.fitness.values = fit
        aggregated_fitness = sum(ind.fitness.wvalues)
        generation_performances.append(aggregated_fitness)
        performances[str(solution_dict["current_generation"])][ind[0].ID] = {
            "agg_fitness": aggregated_fitness,
            "fitness": [float(value) for value in ind.fitness.values],
            "time_stamp": time.perf_counter() - start,
            "hash": ind[0].hash(),
        }

    if optimization.VERBOSE:
        print("Best Performance: ", max(generation_performances))
        print(
            "Average Performance: ",
            sum(generation_performances) / len(generation_performances),
        )


def run_evolutionary_algorithm(
    save_folder: str,
    base_configuration_file_path: str,
    scenario_file_path: str,
    full_save: bool,
    seed: int,
    ngen: int,
    population_size: int,
    mutation_rate: float,
    crossover_rate: float,
    n_seeds: int,
    n_processes: int,
    initial_solutions_folder: str="",
):
    """
    Run an evolutionary algorithm for configuration optimization.

    Args:
        save_folder (str): Folder to save the results in.
        base_configuration_file_path (str): File path of the serialized base configuration (`prodsys.adapters.JsonProductionSystemAdapter`)
        scenario_file_path (str): File path of the serialized scenario (`prodsys.models.scenario_data.ScenarioData`)
        full_save (bool): If True, the full event log of solutions is saved. If False, only the KPIs of solutions are saved.
        seed (int): Random seed for optimization.
        ngen (int): Number of generations to run the algorithm.
        population_size (int): Number of individuals in each generation.
        mutation_rate (float): Probability of mutating an individual.
        crossover_rate (float): Probability of crossover between two individuals.
        n_seeds (int): Number of seeds to use for simulation.
        n_processes (int): Number of processes to use for parallelization.
        initial_solutions_folder (str, optional): If specified, the initial solutions are read from this folder and considered in optimization. Defaults to "".
    """
    base_configuration = adapters.JsonProductionSystemAdapter()
    base_configuration.read_data(base_configuration_file_path, scenario_file_path)

    hyper_parameters = EvolutionaryAlgorithmHyperparameters(
        seed=seed,
        number_of_generations=ngen,
        population_size=population_size,
        mutation_rate=mutation_rate,
        crossover_rate=crossover_rate,
        number_of_processes=n_processes,
        number_of_seeds=n_seeds,
    )

    evolutionary_algorithm_optimization(
        base_configuration,
        hyper_parameters,
        save_folder,
        initial_solutions_folder=initial_solutions_folder,
        full_save=full_save
    )

def optimize_configuration(
    base_configuration_file_path: str,
    scenario_file_path: str,
    save_folder: str,
    hyper_parameters: EvolutionaryAlgorithmHyperparameters,
    full_save: bool = False,
):
    """
    Optimize a configuration using an evolutionary algorithm.

    Args:
        base_configuration_file_path (str): File path of the serialized base configuration (`prodsys.adapters.JsonProductionSystemAdapter`)
        scenario_file_path (str): File path of the serialized scenario (`prodsys.models.scenario_data.ScenarioData`)
        save_folder (str): Folder to save the results in.
        hyper_parameters (EvolutionaryAlgorithmHyperparameters): Hyperparameters for configuration optimization using an evolutionary algorithm.
    """
    run_evolutionary_algorithm(
        save_folder,
        base_configuration_file_path,
        scenario_file_path,
        full_save,
        hyper_parameters.seed,
        hyper_parameters.number_of_generations,
        hyper_parameters.population_size,
        hyper_parameters.mutation_rate,
        hyper_parameters.crossover_rate,
        hyper_parameters.number_of_seeds,
        hyper_parameters.number_of_processes,
    )

from prodsys.util import util

def evolutionary_algorithm_optimization(
        base_configuration: adapters.ProductionSystemAdapter,
        hyper_parameters: EvolutionaryAlgorithmHyperparameters,
        save_folder: str = "results",
        initial_solutions_folder: str = "",
        full_save: bool = False,
):
    """
    Optimize a production system configuration using an evolutionary algorithm.

    Args:
        base_configuration (adapters.ProductionSystemAdapter): production system to optimize.
        hyper_parameters (EvolutionaryAlgorithmHyperparameters): Hyperparameters for configuration optimization using an evolutionary algorithm.
        save_folder (str): Folder to save the results in. Defaults to "results".
        initial_solutions_folder (str, optional): If specified, the initial solutions are read from this folder and considered in optimization. Defaults to "".
    """
    adapters.ProductionSystemAdapter.model_config["validate_assignment"] = False
    base_configuration = base_configuration.model_copy(deep=True)
    if not adapters.check_for_clean_compound_processes(base_configuration):
        logger.warning("Both compound processes and normal processes are used. This may lead to unexpected results.")
    if not check_breakdown_states_available(base_configuration):
        create_default_breakdown_states(base_configuration)

    util.prepare_save_folder(save_folder + "/")
    set_seed(hyper_parameters.seed)

    weights = get_weights(base_configuration, "max")

    # TODO: rework solution_dict and performances to classes
    solution_dict = {
        "current_generation": "0", 
        "hashes": {} 
    }
    performances = {}
    performances["0"] = {}
    start = time.perf_counter()

    toolbox = register_functions_in_toolbox(
        base_configuration=base_configuration,
        solution_dict=solution_dict,
        performances=performances,
        weights=weights,
        initial_solutions_folder=initial_solutions_folder,
        hyper_parameters=hyper_parameters,
        full_save_solutions_folder=save_folder if full_save else "",
    )

    population = toolbox.population(n=hyper_parameters.population_size)
    if hyper_parameters.number_of_processes > 1:
        pool = Pool(hyper_parameters.number_of_processes)
        toolbox.register("map", pool.map)
    else:
        toolbox.register("map", map)

    fitnesses = toolbox.map(toolbox.evaluate, population)
    save_population_results(
        population, fitnesses, solution_dict, performances, save_folder, start
    )

    population = toolbox.select(population, len(population))

    for g in range(hyper_parameters.number_of_generations):
        current_generation = g + 1
        optimization.VERBOSE = True
        if optimization.VERBOSE:
            print(f"\nGeneration: {current_generation}")
        solution_dict["current_generation"] = str(current_generation)
        performances[str(current_generation)] = {}

        # Vary population
        offspring = tools.selTournamentDCD(population, len(population))
        offspring = [toolbox.clone(ind) for ind in offspring]
        offspring = algorithms.varAnd(
            offspring, toolbox, cxpb=hyper_parameters.crossover_rate, mutpb=hyper_parameters.mutation_rate
        )

        # Evaluate the individuals
        fitnesses = toolbox.map(toolbox.evaluate, offspring)
        save_population_results(
            offspring, fitnesses, solution_dict, performances, save_folder, start
        )

        population = toolbox.select(population + offspring, hyper_parameters.population_size)

        with open(f"{save_folder}/optimization_results.json", "w") as json_file:
            json.dump(performances, json_file)
    if hyper_parameters.number_of_processes > 1:
        pool.close()




from typing import Union, Optional
from hashlib import md5

from pydantic import ConfigDict, conlist

from prodsys.models.core_asset import CoreAsset


class QueueData(CoreAsset):
    """
    Class that represents a queue. If capacity is 0, the queue is considered infinite. Otherwise, the queue can hold a finite number of products cooresponding to the capacity.

    Args:
        ID (str): ID of the queue.
        description (str): Description of the queue.
        capacity (Union[int, float]): Capacity of the queue. If 0, the queue is considered infinite. Otherwise, the queue can hold a finite number of products cooresponding to the capacity.

    Examples:
        A finite queue with ID "Q1", description "Queue 1" and capacity 10:
        ``` py
        import prodsys
        prodsys.queue_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
        )
        ```
        An infinite queue with ID "Q1", description "Queue 1" and capacity 0:
        ``` py
        import prodsys
        prodsys.queue_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=0,
        )
        ```
    """

    capacity: Union[int, float] = 0.0
    location: Optional[conlist(float, min_length=2, max_length=2)] = None  # type: ignore

    def hash(self) -> str:
        """
        Returns a unique hash for the queue considering its capacity.


        Returns:
            str: Hash of the queue.
        """
        return md5((str(self.capacity)).encode("utf-8")).hexdigest()

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "ID": "Q1",
                    "description": "Finte Queue",
                    "capacity": 10,
                },
                {
                    "ID": "Q1",
                    "description": "Infinite Queue",
                    "capacity": 0.0,
                },
            ]
        }
    )



from __future__ import annotations

import contextlib
import random
from typing import Any, TYPE_CHECKING

import numpy as np
from simpy import core
from simpy import events


from prodsys.util import util

if util.run_from_ipython():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

# from .factories import state_factory, time_model_factory

if TYPE_CHECKING:
    from prodsys.simulation import request


VERBOSE = 1
"""
Determines whether the simulation should be verbose or not. If set to 1, a progress bar will be shown. Otherwise, no progress bar will be shown.
"""

@contextlib.contextmanager
def temp_seed(seed: int):
    """
    Context manager for temporarily setting the seed of the random number generators. Is necessary when optimizing with another random seed but still wanting to use the same seed for the simulation.

    Args:
        seed (int): The seed to set for the simulation run.
    """
    np_state = np.random.get_state()
    p_state = random.getstate()
    np.random.seed(seed)
    random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(np_state)
        random.setstate(p_state)

class Environment(core.Environment):
    """
    Class to represent the simulation environment. It is a subclass of simpy.Environment and adds a progress bar to the simulation.

    Args:
        seed (int, optional): The seed to set for the simulation run. Defaults to 0.

    Attributes:
        seed (int): The seed to set for the simulation run.
        pbar (Any): The progress bar.
        last_update (int): The last time the progress bar was updated.
    """
    def __init__(self, seed: int=0) -> None:
        super().__init__()
        self.seed: int = seed
        self.pbar: Any = None
        self.last_update = 0

    def run(self, time_range: float):
        """
        Runs the simulation for a given time range.

        Args:
            time_range (int): The time range to run the simulation for in minutes.
        """
        with temp_seed(self.seed):
            if VERBOSE == 1:
                self.pbar = tqdm(total=time_range)

            super().run(time_range)
            if VERBOSE == 1:
                self.pbar.update(time_range - self.last_update)
                self.pbar.close()

    def run_until(self, until: events.Event):
        """
        Runs the simulation until a given event.

        Args:
            until (events.Event): The event to run the simulation until.
        """
        super().run(until=until)

    def request_process_of_resource(self, request: request.Request) -> None:
        """
        Requests the process of a resource. Connects requests of products with controllers in the environment.

        Args:
            request (request.Request): The request to process.
        """
        if VERBOSE == 1:
            now = round(self.now)
            if now > self.last_update:
                self.pbar.update(now - self.last_update)
                self.last_update = now
        controller = request.get_resource().get_controller()
        controller.request(request)



from __future__ import annotations

from abc import ABC
from enum import Enum
from collections.abc import Iterable
from typing import List, Union, Optional, TYPE_CHECKING, Generator

from pydantic import BaseModel, ConfigDict, Field

import logging
logger = logging.getLogger(__name__)

from simpy import events

from pydantic import BaseModel

if TYPE_CHECKING:
    from prodsys.simulation import router, product, resources, sink, source
    from prodsys.factories import auxiliary_factory

from prodsys.models import auxiliary_data
from prodsys.simulation import (
    request,
    process,
    sim,
    store,
    state,
)

class AuxiliaryInfo(BaseModel):
    """
    Class that represents information of the current state of a product.

    Args:
        resource_ID (str): ID of the resource that the product is currently at.
        state_ID (str): ID of the state that the product is currently at.
        event_time (float): Time of the event.
        activity (state.StateEnum): Activity of the product.
        product_ID (str): ID of the product.
        state_type (state.StateTypeEnum): Type of the state.
    """

    resource_ID: str = Field(init=False, default=None)
    state_ID: str = Field(init=False, default=None)
    event_time: float = Field(init=False, default=None)
    activity: state.StateEnum = Field(init=False, default=None)
    product_ID: str = Field(init=False, default=None)
    state_type: state.StateTypeEnum = Field(init=False, default=None)   

    model_config=ConfigDict(extra="allow")         

    def log_create_auxiliary(
        self,
        resource: Union[resources.Resource, sink.Sink, source.Source, store.Queue],
        _product: Auxiliary,
        event_time: float,
    ) -> None:
        """
        Logs the creation of an auxiliary.

        Args:
            resource (Union[resources.Resource, sink.Sink, source.Source]): New resource of the product.
            _product (Product): Product that is created.
            event_time (float): Time of the event.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.data.ID
        self.activity = state.StateEnum.created_auxiliary        
        self.state_type = state.StateTypeEnum.store


    def log_start_auxiliary_usage(
        self,
        resource: resources.Resource,
        _product: Auxiliary,
        event_time: float,
    ) -> None:
        """
        Logs the start of the usage of an auxiliary.

        Args:

            resource (resources.Resource): Resource that the product is processed at.
            _product (Product): Product that is processed.
            event_time (float): Time of the event.
            state_type (state.StateTypeEnum): Type of the state.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.data.ID
        self.activity = state.StateEnum.started_auxiliary_usage
        self.state_type = state.StateTypeEnum.production

    def log_end_auxiliary_usage(
        self,
        resource: resources.Resource,
        _product: Auxiliary,
        event_time: float,
    ) -> None:
        """
        Logs the end of the usage of an auxiliary.

        Args:
            resource (resources.Resource): Resource that the product is processed at.
            _product (Product): Product that is processed.
            event_time (float): Time of the event.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.data.ID
        self.activity = state.StateEnum.finished_auxiliary_usage
        self.state_type = state.StateTypeEnum.production

    def log_start_process(
        self,
        resource: resources.Resource,
        _product: Auxiliary,
        event_time: float,
        state_type: state.StateTypeEnum,
    ) -> None:
        """
        Logs the start of a process.

        Args:
            resource (resources.Resource): Resource that the product is processed at.
            _product (Product): Product that is processed.
            event_time (float): Time of the event.
            state_type (state.StateTypeEnum): Type of the state.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.data.ID
        self.activity = state.StateEnum.start_state
        self.state_type = state_type

    def log_end_process(
        self,
        resource: resources.Resource,
        _product: Auxiliary,
        event_time: float,
        state_type: state.StateTypeEnum,
    ) -> None:
        """
        Logs the end of a process.

        Args:
            resource (resources.Resource): Resource that the product is processed at.
            _product (Product): Product that is processed.
            event_time (float): Time of the event.
            state_type (state.StateTypeEnum): Type of the state.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.data.ID
        self.activity = state.StateEnum.end_state
        self.state_type = state_type

class Auxiliary(BaseModel):
    """
    Class that represents an auxiliary in the discrete event simulation. For easier instantion of the class, use the AuxiliaryFactory at prodsys.factories.auxiliary_factory.

    Args:
        env (sim.Environment): prodsys simulation environment.
        auxilary_data (auxilary.Auxilary): Auxilary data of the product.
    """

    env: sim.Environment
    data: auxiliary_data.AuxiliaryData
    transport_process: process.Process
    storage: store.Queue
    relevant_processes: List[Union[process.ProductionProcess, process.CapabilityProcess]]
    relevant_transport_processes: List[process.TransportProcess]

    
    auxiliary_router: Optional[router.Router] = Field(default=None, init=False)
    current_locatable: Union[product.Locatable, store.Queue] = Field(default=None, init=False)
    current_product: product.Product = Field(default=None, init=False)
    reserved: bool = Field(default=False, init=False)
    got_free: events.Event = Field(default=None, init=False)
    finished_process: events.Event = Field(default=None, init=False)
    auxiliary_info: AuxiliaryInfo = AuxiliaryInfo()


    class Config:
        arbitrary_types_allowed = True

    def init_got_free(self):
        """
        Sets the got_free event.

        Args:
            event (events.Event): The event to set.
        """
        self.got_free = events.Event(self.env)
        self.reserved = False
        self.finished_process = events.Event(self.env)

    @property
    def product_data(self):
        return self.data

    def update_location(self, locatable: product.Locatable):
        """
        Updates the location of the product object.

        Args:
            locatable (Locatable): Location of the product object.
        """
        self.current_locatable = locatable
        logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "resource": self.current_locatable.data.ID, "event": f"Updated location to {self.current_locatable.data.ID}"})

    def reserve(self):
        """
        Reserves the product object.
        """
        self.reserved = True
        self.got_free = events.Event(self.env)
        self.auxiliary_info.log_start_auxiliary_usage(
            resource=self.current_locatable,
            _product=self,
            event_time=self.env.now,
        )
    
    def release_auxiliary_from_product(self):
        """
        Releases the auxiliary from the product after storage of the auxiliary.
        """
        self.current_product = None
        self.reserved = False
        self.got_free.succeed()
        self.auxiliary_info.log_end_auxiliary_usage(
            resource=self.current_locatable,
            _product=self,
            event_time=self.env.now,
        )
        logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "resource": self.current_locatable.data.ID, "event": f"Released auxiliary from product"})


    def request_process(self, processing_request: request.TransportResquest) -> Generator:
        """
        Requests the next production process of the product object from the next production resource by creating a request event and registering it at the environment.
        """
        self.finished_process = events.Event(self.env)

        type_ = state.StateTypeEnum.transport
        logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "resource": processing_request.resource.data.ID, "event": f"Request process {processing_request.process.process_data.ID} for {type_}"})
        self.env.request_process_of_resource(
            request=processing_request
        )
        logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "resource": processing_request.resource.data.ID, "origin": processing_request.origin.data.ID, "target": processing_request.target.data.ID, "event": f"Start waiting for request to be finished"})
        yield self.finished_process
        logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "resource": processing_request.resource.data.ID, "origin": processing_request.origin.data.ID, "target": processing_request.target.data.ID, "event": f"Finished waiting for request to be finished"})
        self.auxiliary_info.log_end_process(
            resource=processing_request.resource,
            _product=self,
            event_time=self.env.now,
            state_type=type_,
        )

from prodsys.simulation import product
